# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import itertools
import pathlib
import sys

import unittest.mock
import typing as t

import pytest_mock

from jinja2.runtime import Context
from jinja2.loaders import DictLoader

import unittest

from ansible._internal import _display_utils
from ansible._internal._templating._datatag import _JinjaConstTemplate
from ansible.errors import (
    AnsibleError, AnsibleUndefinedVariable, AnsibleTemplateSyntaxError, AnsibleBrokenConditionalError, AnsibleTemplateError, AnsibleTemplateTransformLimitError,
    TemplateTrustCheckFailedError,
)
from ansible._internal._templating._errors import AnsibleTemplatePluginRuntimeError, AnsibleTemplatePluginLoadError, AnsibleTemplatePluginNotFoundError
from ansible._internal._errors._handler import ErrorAction, ErrorHandler
from ansible.module_utils._internal._datatag import AnsibleTagHelper, AnsibleDatatagBase
from ansible.module_utils._internal._datatag._tags import Deprecated
from ansible._internal._templating import _transform
from ansible.utils.collection_loader._collection_finder import _AnsibleCollectionFinder
from ansible._internal._datatag._tags import Origin, TrustedAsTemplate
from ansible.plugins.loader import init_plugin_loader
from ansible._internal._templating._jinja_common import _TemplateConfig, _SandboxMode
from ansible._internal._templating._jinja_plugins import _lookup
from ansible._internal._templating import _jinja_plugins
from ansible._internal._templating._engine import TemplateEngine, TemplateOptions
from ansible._internal._templating._jinja_bits import AnsibleEnvironment, AnsibleContext, is_possibly_template, is_possibly_all_template
from ansible._internal._templating._marker_behaviors import ReplacingMarkerBehavior
from ansible._internal._templating._utils import TemplateContext
from ansible.module_utils._internal import _event_utils
from ansible.utils.display import Display
from units.mock.loader import DictDataLoader
from units.test_utils.controller.display import emits_warnings

import pytest

TRUST = TrustedAsTemplate()

origin = Origin(path="/some/path/for/testing", line_num=1, col_num=2)


class BaseTemplar(object):
    def setUp(self):
        init_plugin_loader()
        self.test_vars = dict(
            foo="bar",
            bam=TrustedAsTemplate().tag("{{foo}}"),
            num=1,
            var_true=True,
            var_false=False,
            var_dict=dict(a="b"),
            bad_dict="{a='b'",
            var_list=[1],
            recursive=TrustedAsTemplate().tag("{{recursive}}"),
            some_var="blip",
            some_keyword=TrustedAsTemplate().tag("{{ foo }}"),
            some_unsafe_var="unsafe_blip",
            some_unsafe_keyword=TrustedAsTemplate().tag("{{ foo }}"),
            str_with_error=TrustedAsTemplate().tag("{{ 'str' | from_json }}"),
            template_dict={TrustedAsTemplate().tag("{{ a_keyword }}"): TrustedAsTemplate().tag("{{ some_var }}")},
            template_var=TrustedAsTemplate().tag('{{ some_var }}'),
        )
        self.fake_loader = DictDataLoader({
            "/path/to/my_file.txt": "foo\n",
        })
        self.templar = TemplateEngine(loader=self.fake_loader, variables=self.test_vars)
        self._ansible_context = AnsibleContext(self.templar.environment, {}, {}, {})

    def tearDown(self):
        _AnsibleCollectionFinder._remove()
        nuke_module_prefix('ansible_collections')


class TestTemplarTemplate(BaseTemplar, unittest.TestCase):
    def test_trust_fail_raises_in_tests(self):
        """Ensure template trust check failures default to fatal for unit tests (set in units/conftest.py)"""
        from ansible._internal._templating._engine import TemplateTrustCheckFailedError

        assert _TemplateConfig.untrusted_template_handler.action is ErrorAction.ERROR

        with pytest.raises(TemplateTrustCheckFailedError):
            self.templar.template("{{ i_am_not_trusted }}")

    def test_trust_fail_warning_behavior(self):
        """Validate that trust checks are non-fatal when TemplateConfig.untrusted_template_handler is set to `ErrorAction.WARNING`."""
        untrusted_template = "{{ i_am_not_trusted }}"

        assert hasattr(_TemplateConfig, 'untrusted_template_handler')

        with (unittest.mock.patch.object(_TemplateConfig, 'untrusted_template_handler', ErrorHandler(ErrorAction.WARNING)),
              unittest.mock.patch.object(Display, 'error_as_warning', return_value=None) as mock_warning):
            assert self.templar.template(untrusted_template) is untrusted_template

        assert mock_warning.call_count > 0
        warning_value = mock_warning.call_args.kwargs['exception']
        assert isinstance(warning_value, TemplateTrustCheckFailedError)
        assert "Encountered untrusted template or expression" in warning_value.message
        assert warning_value.obj == untrusted_template

    def test_is_possible_template(self):
        """This test ensures that a broken template still gets templated"""
        # Purposefully invalid jinja
        self.assertRaises(AnsibleError, self.templar.template, TrustedAsTemplate().tag('{{ foo|default(False)) }}'))

    def test_is_template_raw_string(self):
        res = self.templar.is_template('foo')
        self.assertFalse(res)

    def test_is_template_none(self):
        res = self.templar.is_template(None)
        self.assertFalse(res)

    def test_template(self):
        res = self.templar.template(TrustedAsTemplate().tag('{{foo}}'))
        self.assertTrue(res)
        self.assertEqual(res, 'bar')

    def test_template_in_data(self):
        res = self.templar.template(TrustedAsTemplate().tag('{{bam}}'))
        self.assertTrue(res)
        self.assertEqual(res, 'bar')

    def test_template_bare(self):
        res = self.templar.template('bam')
        self.assertTrue(res)
        self.assertEqual(res, 'bam')

    def test_template_to_json(self):
        res = self.templar.template(TrustedAsTemplate().tag('{{bam|to_json}}'))
        self.assertTrue(res)
        self.assertEqual(res, '"bar"')

    def test_template_untagged_string(self):
        unsafe_obj = "Hello"
        res = self.templar.template(unsafe_obj)
        assert not TrustedAsTemplate.is_tagged_on(res)

    def test_weird(self):
        data = TrustedAsTemplate().tag(u"""1 2 #}huh{# %}ddfg{% }}dfdfg{{  {%what%} {{#foo#}} {%{bar}%} {#%blip%#} {{asdfsd%} 3 4 {{foo}} 5 6 7""")
        self.assertRaisesRegex(AnsibleError,
                               'Syntax error in template',
                               self.templar.template,
                               data)

    def test_template_with_error(self):
        """Check that AnsibleError is raised, fail if an unhandled exception is raised"""
        self.assertRaises(AnsibleError, self.templar.template, TrustedAsTemplate().tag("{{ str_with_error }}"))


class TestTemplarMisc(BaseTemplar, unittest.TestCase):
    def test_templar_simple(self):

        templar = self.templar
        # test some basic templating
        self.assertEqual(templar.template(TrustedAsTemplate().tag("{{foo}}")), "bar")
        self.assertEqual(templar.template(TrustedAsTemplate().tag("{{foo}}\n")), "bar\n")
        self.assertEqual(templar.template(TrustedAsTemplate().tag("{{foo}}\n"), options=TemplateOptions(preserve_trailing_newlines=True)), "bar\n")
        self.assertEqual(templar.template(TrustedAsTemplate().tag("{{foo}}\n"), options=TemplateOptions(preserve_trailing_newlines=False)), "bar")
        self.assertEqual(templar.template(TrustedAsTemplate().tag("{{bam}}")), "bar")
        self.assertEqual(templar.template(TrustedAsTemplate().tag("{{num}}")), 1)
        self.assertEqual(templar.template(TrustedAsTemplate().tag("{{var_true}}")), True)
        self.assertEqual(templar.template(TrustedAsTemplate().tag("{{var_false}}")), False)
        self.assertEqual(templar.template(TrustedAsTemplate().tag("{{var_dict}}")), dict(a="b"))
        self.assertEqual(templar.template(TrustedAsTemplate().tag("{{bad_dict}}")), "{a='b'")
        self.assertEqual(templar.template(TrustedAsTemplate().tag("{{var_list}}")), [1])

        # force errors
        self.assertRaises(AnsibleUndefinedVariable, templar.template, TrustedAsTemplate().tag("{{bad_var}}"))
        self.assertRaises(AnsibleUndefinedVariable, templar.template, TrustedAsTemplate().tag("{{lookup('file', bad_var)}}"))
        self.assertRaises(AnsibleError, templar.template, TrustedAsTemplate().tag("{{lookup('bad_lookup')}}"))
        self.assertRaises(AnsibleError, templar.template, TrustedAsTemplate().tag("{{recursive}}"))
        self.assertRaises(AnsibleUndefinedVariable, templar.template, TrustedAsTemplate().tag("{{foo-bar}}"))

        result = templar.extend(marker_behavior=ReplacingMarkerBehavior()).template(TrustedAsTemplate().tag("{{bad_var}}"))
        assert "<< error 1 - 'bad_var' is undefined >>" in result

        # test setting available_variables
        templar.available_variables = dict(foo="bam")
        self.assertEqual(templar.template(TrustedAsTemplate().tag("{{foo}}")), "bam")

    def test_templar_escape_backslashes(self):
        # Rule of thumb: If escape backslashes is True you should end up with
        # the same number of backslashes as when you started.
        self.assertEqual(self.templar.template(TrustedAsTemplate().tag("\t{{foo}}"), options=TemplateOptions(escape_backslashes=True)), "\tbar")
        self.assertEqual(self.templar.template(TrustedAsTemplate().tag("\t{{foo}}"), options=TemplateOptions(escape_backslashes=False)), "\tbar")
        self.assertEqual(self.templar.template(TrustedAsTemplate().tag("\\{{foo}}"), options=TemplateOptions(escape_backslashes=True)), "\\bar")
        self.assertEqual(self.templar.template(TrustedAsTemplate().tag("\\{{foo}}"), options=TemplateOptions(escape_backslashes=False)), "\\bar")
        self.assertEqual(self.templar.template(TrustedAsTemplate().tag("\\{{foo + '\t' }}"), options=TemplateOptions(escape_backslashes=True)), "\\bar\t")
        self.assertEqual(self.templar.template(TrustedAsTemplate().tag("\\{{foo + '\t' }}"), options=TemplateOptions(escape_backslashes=False)), "\\bar\t")
        self.assertEqual(self.templar.template(TrustedAsTemplate().tag("\\{{foo + '\\t' }}"), options=TemplateOptions(escape_backslashes=True)), "\\bar\\t")
        self.assertEqual(self.templar.template(TrustedAsTemplate().tag("\\{{foo + '\\t' }}"), options=TemplateOptions(escape_backslashes=False)), "\\bar\t")
        self.assertEqual(self.templar.template(TrustedAsTemplate().tag("\\{{foo + '\\\\t' }}"), options=TemplateOptions(escape_backslashes=True)), "\\bar\\\\t")
        self.assertEqual(self.templar.template(TrustedAsTemplate().tag("\\{{foo + '\\\\t' }}"), options=TemplateOptions(escape_backslashes=False)), "\\bar\\t")


class TestTemplarLookup(BaseTemplar, unittest.TestCase):
    @staticmethod
    def lookup(name: str, /, *args, **kwargs) -> t.Any:
        with TemplateContext(template_value=None, templar=TemplateEngine(), options=TemplateOptions(), stop_on_template=False):
            return _lookup(name, *args, **kwargs)

    def test_lookup_missing_plugin(self):
        self.assertRaisesRegex(AnsibleTemplatePluginNotFoundError,
                               "The lookup plugin 'not_a_real_lookup_plugin' was not found.",
                               self.lookup,
                               'not_a_real_lookup_plugin',
                               'an_arg', a_keyword_arg='a_keyword_arg_value')

    def test_lookup_list(self):
        res = self.lookup('list', 'an_arg', 'another_arg')
        self.assertEqual(res, 'an_arg,another_arg')

    def test_lookup_jinja_undefined(self):
        self.assertRaisesRegex(AnsibleUndefinedVariable,
                               "'an_undefined_jinja_var' is undefined",
                               self.templar.template,
                               TrustedAsTemplate().tag('{{ lookup("list", an_undefined_jinja_var) }}'))

    def test_lookup_jinja_defined(self):
        res = self.lookup('list', 'x')
        assert not TrustedAsTemplate.is_tagged_on(res)

    def test_lookup_jinja_dict_string_passed(self):
        self.assertRaisesRegex(AnsibleError,
                               "lookup plugin expects a dictionary",
                               self.lookup,
                               'dict',
                               'x')

    def test_lookup_jinja_dict_list_passed(self):
        self.assertRaisesRegex(AnsibleError,
                               "lookup plugin expects a dictionary",
                               self.lookup,
                               'dict',
                               ['foo', 'bar'])

    def test_lookup_jinja_kwargs(self):
        res = self.lookup('list', 'blip', random_keyword='12345')
        assert not TrustedAsTemplate.is_tagged_on(res)

    def test_lookup_jinja_list_wantlist(self):
        res = self.templar.template(TrustedAsTemplate().tag("{{ lookup('list', template_var, wantlist=True) }}"))
        self.assertEqual(res, ["blip"])

    def test_lookup_jinja_list_wantlist_undefined(self):
        self.assertRaisesRegex(AnsibleUndefinedVariable,
                               "'some_undefined_var' is undefined",
                               self.templar.template,
                               TrustedAsTemplate().tag('{{ lookup("list", some_undefined_var, wantlist=True) }}'))

    def test_lookup_jinja_list_wantlist_unsafe(self):
        res = self.lookup('list', 'x', wantlist=True)
        for lookup_result in res:
            assert not TrustedAsTemplate.is_tagged_on(lookup_result)

        assert not TrustedAsTemplate.is_tagged_on(res)

    def test_lookup_jinja_dict(self):
        res = self.templar.template(TrustedAsTemplate().tag('{{ lookup("list", template_dict) }}'))
        self.assertEqual(res['{{ a_keyword }}'], "blip")
        assert not TrustedAsTemplate.is_tagged_on(res)

    def test_lookup_jinja_dict_unsafe(self):
        res = self.lookup('list', {'x': 'x'})
        assert not TrustedAsTemplate.is_tagged_on(res['x'])
        assert not TrustedAsTemplate.is_tagged_on(res)

    def test_lookup_jinja_dict_unsafe_value(self):
        res = self.lookup('list', {'x': 'x'})
        assert not TrustedAsTemplate.is_tagged_on(res['x'])
        assert not TrustedAsTemplate.is_tagged_on(res)

    def test_lookup_jinja_none(self):
        res = self.lookup('list', None)
        self.assertIsNone(res)


class TestAnsibleContext(BaseTemplar, unittest.TestCase):
    def _context(self, variables=None):
        variables = variables or {}

        env = AnsibleEnvironment()
        context = AnsibleContext(env, parent={}, name='some_context',
                                 blocks={})

        for key, value in variables.items():
            context.vars[key] = value

        return context

    def test(self):
        context = self._context()
        self.assertIsInstance(context, AnsibleContext)
        self.assertIsInstance(context, Context)

    def test_resolve_unsafe(self):
        context = self._context(variables={'some_unsafe_key': 'some_unsafe_string'})
        res = context.resolve('some_unsafe_key')
        assert not TrustedAsTemplate.is_tagged_on(res)

    def test_resolve_unsafe_list(self):
        context = self._context(variables={'some_unsafe_key': ['some unsafe string 1']})
        res = context.resolve('some_unsafe_key')
        assert not TrustedAsTemplate.is_tagged_on(res[0])
        assert not TrustedAsTemplate.is_tagged_on(res)

    def test_resolve_unsafe_dict(self):
        context = self._context(variables={'some_unsafe_key':
                                           {'an_unsafe_dict': 'some unsafe string 1'}
                                           })
        res = context.resolve('some_unsafe_key')
        assert not TrustedAsTemplate.is_tagged_on(res['an_unsafe_dict'])

    def test_resolve(self):
        context = self._context(variables={'some_key': 'some_string'})
        res = context.resolve('some_key')
        self.assertEqual(res, 'some_string')

    def test_resolve_none(self):
        context = self._context(variables={'some_key': None})
        res = context.resolve('some_key')
        self.assertEqual(res, None)


def test_unsafe_lookup():
    res = TemplateEngine(
        None,
        variables={
            'var0': TrustedAsTemplate().tag('{{ var1 }}'),
            'var1': ['unsafe'],
        }
    ).template(TrustedAsTemplate().tag('{{ lookup("list", var0) }}'))
    assert not TrustedAsTemplate.is_tagged_on(res[0])


def test_unsafe_lookup_no_conversion():
    res = TemplateEngine(
        None,
        variables={
            'var0': TrustedAsTemplate().tag('{{ var1 }}'),
            'var1': ['unsafe'],
        }
    ).template(
        TrustedAsTemplate().tag('{{ lookup("list", var0) }}'),
    )
    assert not TrustedAsTemplate.is_tagged_on(res)


@pytest.mark.parametrize("tagged", (
    False,
    True,
))
def test_dict_template(tagged: bool) -> None:
    """Verify that templar.template can round-trip both tagged and untagged values in a dict."""
    key1 = "key1"
    val1 = "val1"

    if tagged:
        key1 = origin.tag(key1)
        val1 = origin.tag(val1)

    test1 = {
        key1: val1,
    }

    variables = dict(
        test1=test1,
    )

    templar = TemplateEngine(variables=variables)

    result = templar.template(TrustedAsTemplate().tag('{{test1}}'))

    assert result == test1
    assert AnsibleTagHelper.tags(result) == AnsibleTagHelper.tags(test1)


@pytest.mark.parametrize("expr,expected,variables", [
    ("'constant'", "constant", None),
    ("a - b", 42, dict(a=100, b=58)),
])
def test_evaluate_expression(expr: str, expected: t.Any, variables: dict[str, t.Any] | None):
    assert TemplateEngine(variables=variables).evaluate_expression(TRUST.tag(expr)) == expected


@pytest.mark.parametrize("expr,error_type", [
    ("fhdgsfk#$76&@#$&", AnsibleTemplateSyntaxError),
    ("bogusvar", AnsibleUndefinedVariable),
    ("untrusted expression", TemplateTrustCheckFailedError),
    (dict(hi="{{'mom'}}"), TypeError),
])
def test_evaluate_expression_errors(expr: str, error_type: type[Exception]):
    if error_type is not TemplateTrustCheckFailedError:
        expr = TRUST.tag(expr)

    with pytest.raises(error_type):
        TemplateEngine().evaluate_expression(expr)


@pytest.mark.parametrize("conditional,expected,variables", [
    ("1 == 2", False, None),
    ("test2_name | default(True)", True, None),
    # DTFIX5: more success cases?
])
def test_evaluate_conditional(conditional: str, expected: t.Any, variables: dict[str, t.Any] | None):
    assert TemplateEngine().evaluate_conditional(TRUST.tag(conditional)) == expected


@pytest.mark.parametrize("conditional,error_type", [
    ("fkjhs$#@^%$*& ldfkjds", AnsibleTemplateSyntaxError),
    ("#jinja2:variable_start_string:2\n{{blah}}", AnsibleTemplateSyntaxError),
    ("#jinja2:bogus_key:'val'\n{{blah}}", AnsibleTemplateSyntaxError),
    ("bogusvar", AnsibleUndefinedVariable),
    ("not trusted", TemplateTrustCheckFailedError),
])
def test_evaluate_conditional_errors(conditional: t.Any, error_type: type[Exception], mocker: pytest_mock.MockerFixture):
    mocker.patch.object(_TemplateConfig, 'allow_embedded_templates', True)  # force this on since a number of cases need it

    if error_type is not TemplateTrustCheckFailedError:
        conditional = TRUST.tag(conditional)

    with pytest.raises(error_type):
        TemplateEngine().evaluate_conditional(conditional)


@pytest.mark.parametrize("value", (
    '{{ foo }}',
    '{% foo %}',
    '{# foo #}',
    '{# {{ foo }} #}',
    '{# {{ nothing }} {# #}',
    '{# {{ nothing }} {# #} #}',
    '{% raw %}{{ foo }}{% endraw %}',
    # in 2.16 and earlier these were not considered templates due to syntax errors
    # now syntax errors in templates are still reported as templates, since is_template no longer compiles the template
    '{{ foo',
    '{% foo',
    '{# foo',
    '{{ foo %}',
    '{{ foo #}',
    '{% foo }}',
    '{% foo #}',
    '{# foo %}',
    '{# foo }}',
    '{{ foo {{',
    '{% raw %}{% foo %}',
))
def test_is_template_true(value: str) -> None:
    assert TemplateEngine().is_template(TRUST.tag(value))


@pytest.mark.parametrize("value", (
    'foo',
))
def test_is_template_false(value: str) -> None:
    assert not TemplateEngine().is_template(TRUST.tag(value))


@pytest.mark.parametrize("value", (
    '{{ foo }}',
    '{% foo %}',
    '{# foo #}',
    '{# {{ foo }} #}',
    '{# {{ nothing }} {# #}',
    '{# {{ nothing }} {# #} #}',
    '{% raw %}{{ foo }}{% endraw %}',
    '{{',
    '{%',
    '{#',
    '{% raw',
))
def test_is_possibly_template_true(value: str) -> None:
    assert is_possibly_template(value)


@pytest.mark.parametrize("value", (
    '{',
    '%',
    '#',
    'foo',
    '}}',
    '%}',
    'raw %}',
    '#}',
))
def test_is_possibly_template_false(value: str) -> None:
    assert not is_possibly_template(value)


def test_stop_on_container() -> None:
    # DTFIX5: add more test cases
    assert TemplateEngine().resolve_to_container(TRUST.tag('{{ [ 1 ] }}')) == [1]


@pytest.mark.parametrize("value", [True, False])
def test_stripped_conditionals(value: bool, mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch.object(_TemplateConfig, 'allow_embedded_templates', True)  # force this on since this case needs it

    assert TemplateEngine().evaluate_conditional(TRUST.tag(f"""\n \r\n \t{{{{ {value} }}}} \n\n  \t \t\t  """)) == value


@pytest.mark.parametrize("template, variables, error", (
    ("{{ undefined_var.undefined_attribute }}", {}, "'undefined_var' is undefined >>"),
    ("{{ some_dict['undefined_key'] }}", dict(some_dict={}), "object of type 'dict' has no attribute 'undefined_key' >>"),
    ("{{ some_dict.undefined_key }}", dict(some_dict={}), "object of type 'dict' has no attribute 'undefined_key' >>"),
    ("{{ m1 }} {{ m2 }} here", {}, "<< error 1 - 'm1' is undefined >> << error 2 - 'm2' is undefined >> here"),
    ("before {{ m1 + m2 }} after", {}, "before << error 1 - 'm1' is undefined >><< error 2 - template potentially truncated >>"),
))
def test_jinja_sourced_undefined(template: str, variables: dict[str, t.Any], error: str) -> None:
    """
    Ensure when Jinja encounters a `Marker` and raises `MarkerError`,
    that we turn it back into the original `Marker` so marker_behavior can handle it during finalization.
    """
    assert error in TemplateEngine(variables=variables, marker_behavior=ReplacingMarkerBehavior()).template(TRUST.tag(template))


def test_omit_concat() -> None:
    assert TemplateEngine().template(TRUST.tag("{{ omit }}hi{{ omit }} mom")) == 'hi mom'


@pytest.mark.parametrize("conditional", (
    # Jinja plugins
    "'join' is filter",
    "'join' is not test",
    "'eq' is test",
    "'eq' is not filter",

    # Ansible plugins
    "'comment' is filter",
    "'comment' is not test",
    "'version' is test",
    "'version' is not filter",

    # plugin not found
    "'nope' is not filter",
    "'nope' is not test",
))
def test_plugin_found_not_found(conditional: str) -> None:
    assert TemplateEngine().evaluate_conditional(TRUST.tag(conditional))


@pytest.mark.parametrize("value, expected", (
    ("{{ {'a': 1}.items() }}", [['a', 1]]),
    ("{{ {'a': 1}.keys() }}", ['a']),
    ("{{ {'a': 1}.values() }}", [1]),
    ("{{ yielder(2) }}", [0, 1]),
    ("{% set y = yielder(2) %}{{ y | list }} | {{ y | list }}", "[0, 1] | [0, 1]"),
), ids=str)
def test_finalize_generator(value: t.Any, expected: t.Any) -> None:
    def yielder(count: int) -> t.Generator[int, None, None]:
        yield from range(count)

    templar = TemplateEngine(variables=dict(
        yielder=yielder,
    ))

    # DTFIX5: we still need to deal with the "Encountered unsupported" warnings these generate
    assert templar.template(TRUST.tag(value)) == expected


@pytest.mark.parametrize("template", (
    "{{ lookup('my_lookup', some_var) }}",
    "{{ some_var | my_filter }}",
    "{{ some_var is my_test }}",
))
def test_eager_trip_undefined(template: str, mocker: pytest_mock.MockerFixture) -> None:
    """Verify that eager tripping of Marker works for template plugins which only perform isinstance checks on undefined values."""
    from ansible.plugins.lookup import LookupBase

    class MyLookup(LookupBase):
        def run(self, terms, variables=None, **kwargs):
            return [isinstance(value, bool) for value in itertools.chain(*terms)]

    def my_filter(values):
        return [isinstance(value, bool) for value in values]

    def my_test(values):
        return all(isinstance(value, bool) for value in values)

    def mock_lookup_get(*_args, **_kwargs) -> t.Any:
        return MyLookup()

    mock_lookup_loader = mocker.MagicMock()
    mock_lookup_loader.get = mock_lookup_get

    mocker.patch.object(_jinja_plugins, 'lookup_loader', mock_lookup_loader)

    def get_templar(variables: dict[str, t.Any]) -> TemplateEngine:
        new_templar = TemplateEngine(variables=variables)
        new_templar.environment.filters['my_filter'] = my_filter
        new_templar.environment.tests['my_test'] = my_test

        return new_templar

    template = TRUST.tag(template)

    # verify the template works when some_var is defined

    templar = get_templar(dict(some_var=[True]))

    result = templar.template(template)

    assert result is True or result == [True]

    # verify the template raises AnsibleUndefinedVariable when some_var contains a template that references an undefined variable

    templar = get_templar(dict(some_var=[TRUST.tag("{{ nope }}")]))

    with pytest.raises(AnsibleUndefinedVariable) as ex:
        templar.template(template)

    assert ex.value.message == "'nope' is undefined"


def as_template(value: str) -> str:
    return f"{{{{ {value} }}}}"


TEMPLATED_LOOKUP_NAME_TEST_VALUES = [
    ("""lookup('{{ "pipe" }}', 'echo hi')""", "hi"),
    ("""query('{{ "pipe" }}', 'echo hi')""", ["hi"]),
]


@pytest.mark.parametrize("value", [v[0] for v in TEMPLATED_LOOKUP_NAME_TEST_VALUES])
def test_lookup_query_name_is_not_templated_non_conditional(value: str) -> None:
    with pytest.raises(AnsibleTemplatePluginNotFoundError):
        TemplateEngine().template(TRUST.tag(as_template(value)))


@pytest.mark.parametrize("value", [v[0] for v in TEMPLATED_LOOKUP_NAME_TEST_VALUES])
def test_lookup_query_name_is_not_templated_conditional_nested_template(value: str, mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch.object(_TemplateConfig, 'allow_embedded_templates', True)  # force this on since a number of cases need it

    with pytest.raises(AnsibleTemplatePluginNotFoundError):
        TemplateEngine().evaluate_conditional(TRUST.tag(as_template(value)))


@pytest.mark.parametrize("value, expected_result", TEMPLATED_LOOKUP_NAME_TEST_VALUES)
def test_lookup_query_name_is_not_templated_conditional_expression(value: str, expected_result: t.Any, mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch.object(_TemplateConfig, 'allow_embedded_templates', True)  # force this on since a number of cases need it

    with emits_warnings(warning_pattern="should not contain embedded templates"):
        assert TemplateEngine().evaluate_conditional(TRUST.tag(f'{value} == {expected_result!r}'))


@pytest.mark.parametrize("value", [
    "foo(",
    "'a' == {{ 'b' }}",
])
def test_conditional_syntax_error(value: str) -> None:
    with pytest.raises(AnsibleTemplateSyntaxError):
        TemplateEngine().evaluate_conditional(TRUST.tag(value))


BROKEN_CONDITIONAL_VALUES = [
    (None, True),  # stupid backward-compat
    ("", True),  # stupid backward-compat
    ("''", False),
    ("0", False),
    ("0.0", False),
    ("1", True),
    ("1.1", True),
    ("'abc'", True),
    ("{{ '' }}", True),
    ("{{ None }}", True),
    ("{{ 0 }}", False),
    ("{{ 0.0 }}", False),
    ("{{ [] }}", False),
    ("{{ {} }}", False),
    ([], False),
    ([TRUST.tag("{{ omit }}")], False),
    ({}, False),
    (dict(a=TRUST.tag("{{ omit }}")), False),
    (["abc", TRUST.tag("{{ omit }}")], True),
    (dict(a="b", omitted=TRUST.tag("{{ omit }}")), True),
    (0, False),
    (0.0, False),
    (1, True),
    (1.1, True),
]


@pytest.mark.parametrize("value", [v[0] for v in BROKEN_CONDITIONAL_VALUES], ids=repr)
def test_broken_conditionals_disabled(value: t.Any, mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch.object(_TemplateConfig, 'allow_broken_conditionals', False)

    with pytest.raises(AnsibleBrokenConditionalError):
        TemplateEngine().evaluate_conditional(TRUST.tag(value))


@pytest.mark.parametrize("value, expected_result", BROKEN_CONDITIONAL_VALUES, ids=repr)
def test_broken_conditionals_enabled(value: t.Any, expected_result: bool, mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch.object(_TemplateConfig, 'allow_broken_conditionals', True)
    mocker.patch.object(_TemplateConfig, 'allow_embedded_templates', True)  # force this on since a number of cases need it

    deprecation_matches = []

    if isinstance(value, str) and is_possibly_all_template(value):
        deprecation_matches.append("should not be surrounded")

    if value in (None, '', "{{ '' }}", "{{ None }}"):
        deprecation_matches.append("Empty conditional")
    else:
        deprecation_matches.append("must have a boolean result")

    with emits_warnings(deprecation_pattern=deprecation_matches):
        assert TemplateEngine().evaluate_conditional(TRUST.tag(value)) == expected_result


@pytest.mark.parametrize("template, expected", (
    ("1 == '{{ 1 }}'", False),
    ("lookup('items', '{{ [1, 2, 3] }}') == [1, 2, 3]", False),
    ("query('items', '{{ [1, 2, 3] }}') == [1, 2, 3]", False),
))
def test_embedded_templates_disabled(template: str, expected: t.Any, mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch.object(_TemplateConfig, 'allow_embedded_templates', False)

    with emits_warnings(warning_pattern=[]):
        assert TemplateEngine().evaluate_conditional(TRUST.tag(template)) == expected


@pytest.mark.parametrize("template, expected", (
    ("1 == '{{ 1 }}'", True),
    ("lookup('items', '{{ [1, 2, 3] }}') == [1, 2, 3]", True),
    ("query('items', '{{ [1, 2, 3] }}') == [1, 2, 3]", True),
))
def test_embedded_templates_enabled(template: str, expected: t.Any, mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch.object(_TemplateConfig, 'allow_embedded_templates', True)

    templar = TemplateEngine()

    with emits_warnings(warning_pattern="should not contain embedded templates"):
        assert templar.evaluate_conditional(TRUST.tag(template)) == expected

    # only lookup/query args support embedded templates in an actual template (not a naked expression)
    if 'lookup' in template or 'query' in template:
        with emits_warnings(warning_pattern="should not contain embedded templates"):
            assert templar.evaluate_conditional(TRUST.tag(as_template(template))) == expected


def test_available_vars_smuggling():
    """
    Jinja Template.render() and TemplateExpression.__call__() flatten their args/kwargs via splatting to dict(), which is an unnecessary copy as well as
    causing top-level templated variables to be rendered prematurely. We have some arg smuggling code to prevent this, which this test validates.
    """

    class ExplodingDict(dict):
        """A dict subclass that explodes when copied or iterated via `dict()`."""

        def __iter__(self):
            raise NotImplementedError()

        def keys(self):
            raise NotImplementedError()

    template_vars = ExplodingDict()

    # ensure our tripwire dict subclass fails when used as the source for a dict copy
    with pytest.raises(NotImplementedError):
        dict(template_vars)

    # if Jinja copies our input dict, it should blow up; assert that it doesn't and that the template renders as expected
    assert TemplateEngine(variables=template_vars).template(TRUST.tag("{{ 1 }}")) == 1


def test_template_var_isolation():
    """
    Ensure that plugin mutations to lazy-wrapped container variables do not persist outside templating.
    Direct mutation via Templar.available_variables is not currently protected (and is generally a bad idea).
    """

    orig_dict_value = dict(one="one")
    orig_list_value = [1, 2, 3]

    available_vars = dict(dict_value=orig_dict_value, list_value=orig_list_value)

    def mutate_my_vars(dict_value: dict, list_value: list) -> dict[str, t.Any]:
        dict_value["added"] = "added"
        list_value.append("added")

        return dict(dict_value=dict_value, list_value=list_value)

    res = TemplateEngine(variables=available_vars).evaluate_expression(
        TRUST.tag("mutate_my_vars(dict_value, list_value)"),
        local_variables=dict(mutate_my_vars=mutate_my_vars)
    )

    # ensure the plugin returned the mutated copies as expected
    assert res == dict(dict_value=dict(one="one", added="added"), list_value=[1, 2, 3, "added"])

    # ensure the original input variables are the same unmodified instances
    assert available_vars == dict(dict_value=dict(one="one"), list_value=[1, 2, 3])
    assert available_vars['dict_value'] is orig_dict_value
    assert available_vars['list_value'] is orig_list_value


@pytest.mark.parametrize('fixture, plugin_type, plugin_name, expected', (
    ('no_collections', 'filter', 'invalid/name.does_not_matter.also_does_not_matter', AnsibleTemplateSyntaxError),  # plugin is None
    ('no_collections', 'lookup', 'invalid/name.does_not_matter.also_does_not_matter', AnsibleTemplatePluginNotFoundError),
    ('no_collections', 'filter', 'missing_namespace_name.does_not_matter.also_does_not_matter', AnsibleTemplateSyntaxError),  # KeyError
    ('no_collections', 'lookup', 'missing_namespace_name.does_not_matter.also_does_not_matter', AnsibleTemplatePluginNotFoundError),

    ('valid_collection', 'filter', 'valid.invalid/name.does_not_matter', AnsibleTemplateSyntaxError),  # KeyError
    ('valid_collection', 'lookup', 'valid.invalid/name.does_not_matter', AnsibleTemplatePluginNotFoundError),
    ('valid_collection', 'filter', 'valid.missing_collection.does_not_matter', AnsibleTemplateSyntaxError),  # KeyError
    ('valid_collection', 'lookup', 'valid.missing_collection.does_not_matter', AnsibleTemplatePluginNotFoundError),
    ('valid_collection', 'filter', 'valid.also_valid.invalid/name', AnsibleTemplateSyntaxError),  # plugin is None
    ('valid_collection', 'lookup', 'valid.also_valid.invalid/name', AnsibleTemplatePluginNotFoundError),
    ('valid_collection', 'filter', 'valid.also_valid.missing_plugin', AnsibleTemplateSyntaxError),  # plugin is None
    ('valid_collection', 'lookup', 'valid.also_valid.missing_plugin', AnsibleTemplatePluginNotFoundError),
    ('valid_collection', 'filter', 'valid.also_valid.also_also_valid', []),
    ('valid_collection', 'lookup', 'valid.also_valid.also_also_valid', []),
    ('valid_collection', 'filter', 'valid.also_valid.runtime_error', AnsibleTemplatePluginRuntimeError),
    ('valid_collection', 'lookup', 'valid.also_valid.runtime_error', AnsibleTemplatePluginRuntimeError),
    ('valid_collection', 'filter', 'valid.also_valid.load_error', AnsibleTemplatePluginLoadError),  # AnsibleError
    ('valid_collection', 'lookup', 'valid.also_valid.load_error', AnsibleTemplatePluginLoadError),
    ('valid_collection', 'template', '{% if false %}{{ 123 | valid.also_valid.load_error }}{% else %}Success{% endif %}', AnsibleTemplatePluginLoadError),

    ('no_collections', 'filter', 'ansible.invalid/name.does_not_matter', AnsibleTemplateSyntaxError),  # KeyError
    ('no_collections', 'lookup', 'ansible.invalid/name.does_not_matter', AnsibleTemplatePluginNotFoundError),
    ('no_collections', 'filter', 'ansible.missing_collection.does_not_matter', AnsibleTemplateSyntaxError),  # KeyError
    ('no_collections', 'lookup', 'ansible.missing_collection.does_not_matter', AnsibleTemplatePluginNotFoundError),
    ('no_collections', 'filter', 'ansible.builtin.invalid/name', AnsibleTemplateSyntaxError),  # plugin is None
    ('no_collections', 'lookup', 'ansible.builtin.invalid/name', AnsibleTemplatePluginNotFoundError),
    ('no_collections', 'filter', 'ansible.builtin.missing_plugin', AnsibleTemplateSyntaxError),  # plugin is None
    ('no_collections', 'lookup', 'ansible.builtin.missing_plugin', AnsibleTemplatePluginNotFoundError),
    ('no_collections', 'filter', 'ansible.builtin.quote', 'foo'),
    ('no_collections', 'lookup', 'ansible.builtin.env', []),
    ('no_collections', 'template', '{% if false %}{{ 123 | ansible.builtin.missing_plugin }}{% else %}Success{% endif %}', 'Success'),

    ('no_collections', 'filter', 'invalid/name', AnsibleTemplateSyntaxError),  # plugin is None
    ('no_collections', 'lookup', 'invalid/name', AnsibleTemplatePluginNotFoundError),
    ('no_collections', 'filter', 'missing_plugin', AnsibleTemplateSyntaxError),  # plugin is None
    ('no_collections', 'lookup', 'missing_plugin', AnsibleTemplatePluginNotFoundError),
    ('no_collections', 'filter', 'quote', 'foo'),
    ('no_collections', 'lookup', 'env', []),
), ids=str)
def test_jinja2_loader_plugin(fixture: str, plugin_type: str, plugin_name: str, expected: t.Any) -> None:
    if plugin_type == 'filter':
        expression = f'{{{{ "foo" | {plugin_name} }}}}'
    elif plugin_type == 'template':
        expression = plugin_name  # abusing plugin_type and plugin_name to allow for testing arbitrary expressions
    else:
        expression = f'{{{{ lookup("{plugin_name}") }}}}'

    # HACK: this test should really be using a shared collection loader fixture, but this fixes an "inherited" dummy collection loader
    from ansible.utils.collection_loader._collection_finder import _AnsibleCollectionFinder

    try:
        _AnsibleCollectionFinder._remove()
        nuke_module_prefix('ansible_collections')
    except Exception:
        pass

    _AnsibleCollectionFinder(paths=[str(pathlib.Path(__file__).parent / 'fixtures' / fixture)])._install()

    try:
        if isinstance(expected, type) and issubclass(expected, Exception):
            with pytest.raises(expected):
                TemplateEngine().template(TRUST.tag(expression))
        else:
            assert TemplateEngine().template(TRUST.tag(expression)) == expected
    finally:
        _AnsibleCollectionFinder._remove()

        nuke_module_prefix('ansible_collections')


def test_variable_name_as_template_success() -> None:
    name = origin.tag("blar")

    res = TemplateEngine().variable_name_as_template(name)
    assert res.replace(' ', '') == "{{blar}}"

    required_tags: frozenset[AnsibleDatatagBase] = frozenset({origin, TrustedAsTemplate()})

    assert required_tags - AnsibleTagHelper.tags(res) == set()  # there might be others, that's fine


def test_variable_name_as_template_invalid() -> None:
    invalid_name = origin.tag("  invalid[var*name")

    with pytest.raises(AnsibleError) as err:
        TemplateEngine().variable_name_as_template(invalid_name)

    assert err.value.obj is invalid_name


@pytest.mark.parametrize("expression, expected", (
    ("dictthing.subdict1.subdict2", "hi mom"),
    ("dictthing.sublist1[0]", 1),
    ("dictthing[keys.sublist1][0]", 1),
    ("123", 123),
))
def test_resolve_variable_expression_success(expression: str, expected: t.Any) -> None:
    templar = TemplateEngine()

    local_variables = dict(
        dictthing=dict(sublist1=[1, 2, 3], subdict1=dict(subdict2="hi mom")),
        keys=dict(sublist1="sublist1")
    )

    assert templar.resolve_variable_expression(expression, local_variables=local_variables) == expected


@pytest.mark.parametrize("expression", (
    "'text'",
    "dictthing['subdict1']",
    "q('env')",
))
def test_resolve_variable_expression_invalid(expression: str) -> None:
    templar = TemplateEngine()

    with pytest.raises(AnsibleError) as err:
        templar.resolve_variable_expression(expression, local_variables=dict(dictthing=dict(subdict1=1)))

    assert err.value.obj is expression


def test_resolve_variable_expression_missing() -> None:
    templar = TemplateEngine()

    expr = origin.tag("missing_variable")

    with pytest.raises(AnsibleUndefinedVariable) as err:
        templar.resolve_variable_expression(expr)

    assert err.value.obj == expr  # may not be the same instance, since it was tagged TrustedAsTemplate internally

    required_tags: frozenset[AnsibleDatatagBase] = frozenset({origin, TrustedAsTemplate()})

    assert required_tags - AnsibleTagHelper.tags(err.value.obj) == set()  # there might be others, that's fine


def test_error_invalid_non_string_template():
    """Ensure errors on non-string template inputs include type information."""
    with pytest.raises(AnsibleTemplateError) as err:
        TemplateEngine().template(...)

    assert f"Type {type(...).__name__!r} is unsupported for variable storage." in err.value.message


def nuke_module_prefix(prefix):
    for module_to_nuke in [m for m in sys.modules if m.startswith(prefix)]:
        sys.modules.pop(module_to_nuke)


def test_template_transform_limit_exceeded(mocker: pytest_mock.MockerFixture) -> None:
    """
    Verify that template transforms cannot trigger an infinite loop.
    This currently requires injecting bogus transforms to trigger the condition, but the logic is present to catch future coding errors.
    """
    class One:
        def __init__(self, *args, **kwargs):
            pass

    class Two:
        def __init__(self, *args, **kwargs):
            pass

    mocker.patch.dict(_transform._type_transform_mapping, {One: Two, Two: One})

    with pytest.raises(AnsibleTemplateTransformLimitError):
        TemplateEngine(variables=dict(limit=One())).template(TRUST.tag("{{ limit }}"))


def test_transform_transform_limit_exceeded(mocker: pytest_mock.MockerFixture) -> None:
    """
    Verify that standalone recursive transforms cannot trigger an infinite loop.
    This currently requires injecting bogus transforms to trigger the condition, but the logic is present to catch future coding errors.
    """
    class One:
        def __init__(self, *args, **kwargs):
            pass

    class Two:
        def __init__(self, *args, **kwargs):
            pass

    mocker.patch.dict(_transform._type_transform_mapping, {One: Two, Two: One})

    with pytest.raises(AnsibleTemplateTransformLimitError):
        TemplateEngine().transform(One())


def test_deprecated_dedupe_and_source():
    """Validate dedupe and source context behavior for deprecated item access and associated warning behavior."""
    # unique tag instances that share the same contents (can be tracked independently by the audit context)
    deprecated_string = Deprecated(msg="deprecated").tag("deprecated string")
    deprecated_list = Deprecated(msg="deprecated").tag([42])
    deprecated_dict = Deprecated(msg="deprecated").tag(dict(key="value"))

    # a shared tag instance (cannot be tracked independently by the audit context)
    shared_tag_instance = Deprecated(msg="shared tag")
    d1 = shared_tag_instance.tag("d1")
    d2 = shared_tag_instance.tag("d2")

    variables = dict(
        indirect1=TRUST.tag('{{ indirect2 }}'),
        indirect2=TRUST.tag('{{ deprecated_string }}'),
        deprecated_string=deprecated_string,
        deprecated_list=deprecated_list,
        deprecated_dict=deprecated_dict,
        d1=d1,
        d2=d2,
        ansible_deprecation_warnings=True,
    )

    templar = TemplateEngine(variables=variables)

    with _display_utils.DeferredWarningContext(variables=variables) as dwc:
        # The indirect access summary occurs first.
        # The two following direct access summaries get deduped to a single one by the warning context (but unique template value keeps distinct from indirect).
        # The accesses with the shared tag instance values are internally deduped by the audit context.
        templar.evaluate_expression(TRUST.tag("indirect1 and deprecated_list and deprecated_dict and d1 and d2"))

    dep_warnings = dwc.get_deprecation_warnings()

    assert len(dep_warnings) == 3
    assert 'deprecated_string' in _event_utils.format_event_brief_message(dep_warnings[0].event)
    assert 'indirect1 and deprecated_list and deprecated_dict' in _event_utils.format_event_brief_message(dep_warnings[1].event)
    assert 'd1 and d2' in _event_utils.format_event_brief_message(dep_warnings[2].event)


def test_jinja_const_template_leak(template_context: TemplateContext) -> None:
    """Verify that _JinjaConstTemplate is present during internal templating."""
    with _display_utils.DeferredWarningContext(variables={}):  # suppress warning from usage of embedded template
        with unittest.mock.patch.object(_TemplateConfig, 'allow_embedded_templates', True):
            assert _JinjaConstTemplate.is_tagged_on(TemplateEngine().template(TRUST.tag("{{ '{{ 1 }}' }}")))


def test_jinja_const_template_finalized() -> None:
    """Verify that _JinjaConstTemplate is not present in finalized template results."""
    with _display_utils.DeferredWarningContext(variables={}):  # suppress warning from usage of embedded template
        with unittest.mock.patch.object(_TemplateConfig, 'allow_embedded_templates', True):
            assert not _JinjaConstTemplate.is_tagged_on(TemplateEngine().template(TRUST.tag("{{ '{{ 1 }}' }}")))


@pytest.mark.parametrize("template,expected", (
    ("{{ (-1).__abs__() }}", 1),
))
def test_unsafe_attr_access(template: str, expected: object) -> None:
    """Verify that unsafe attribute access fails by default and works when explicitly configured."""
    assert _TemplateConfig.sandbox_mode == _SandboxMode.DEFAULT

    with pytest.raises(AnsibleUndefinedVariable):
        TemplateEngine().template(TRUST.tag(template))

    with unittest.mock.patch.object(_TemplateConfig, 'sandbox_mode', _SandboxMode.ALLOW_UNSAFE_ATTRIBUTES):
        assert TemplateEngine().template(TRUST.tag(template)) == expected


def test_marker_from_test_plugin() -> None:
    """Verify test plugins can raise MarkerError to return a Marker, and that no warnings or deprecations are emitted."""
    with emits_warnings(deprecation_pattern=[], warning_pattern=[]):
        assert TemplateEngine(variables=dict(something=TRUST.tag("{{ nope }}"))).template(TRUST.tag("{{ (something is eq {}) is undefined }}"))


@pytest.mark.parametrize("template,expected", (
    ("{{ none }}", None),  # concat sees one node, NoneType result is preserved
    ("{% if False %}{% endif %}", None),  # concat sees one node, NoneType result is preserved
    ("{{''}}{% if False %}{% endif %}", ""),  # multiple blocks with an embedded None result, concat is in play, the result is an empty string
    ("hey {{ none }}", "hey "),  # composite template, the result is an empty string
    ("{% import 'importme' as imported %}{{ imported }}", "imported template result"),
))
def test_none_concat(template: str, expected: object) -> None:
    """Validate that None values are omitted from composite template concat."""
    te = TemplateEngine()

    # set up an importable template to exercise TemplateModule code paths
    te.environment.loader = DictLoader(dict(importme=TRUST.tag("{{ none }}{{ 'imported template result' }}{{ none }}")))

    assert te.template(TRUST.tag(template)) == expected


def test_filter_generator() -> None:
    """Verify that filters which return a generator are converted to a list while under the filter's JinjaCallContext."""
    variables = dict(
        foo=[
            dict(x=1, optional_var=0),
            dict(x=2),
        ],
        bar=TRUST.tag("{{ foo | selectattr('optional_var', 'defined') }}"),
    )

    te = TemplateEngine(variables=variables)
    te.template(TRUST.tag("{{ bar }}"))
    te.template(TRUST.tag("{{ lookup('vars', 'bar') }}"))


def test_call_context_reset() -> None:
    """Ensure that new template invocations do not inherit trip behavior from running Jinja plugins."""
    templar = TemplateEngine(variables=dict(
        somevar=TRUST.tag("{{ somedict.somekey | default('ok') }}"),
        somedict=dict(
            somekey=TRUST.tag("{{ not_here }}"),
        )
    ))

    assert templar.template(TRUST.tag("{{ lookup('vars', 'somevar') }}")) == 'ok'



from __future__ import annotations

import functools

from ansible.plugins.callback import CallbackBase


class CallbackModule(CallbackBase):
    """Test callback that implements exclusively deprecated v1 callback methods."""
    CALLBACK_NEEDS_ENABLED = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.called_v1_method_names: set[str] = set()

    def callback_impl(self, *args, name: str, **kwargs) -> None:
        print(f"hi from callback {name!r} with {args=!r} {kwargs=!r}")
        self.called_v1_method_names.add(name)

    for v1_method in CallbackBase._v2_v1_method_map.values():
        if not v1_method:
            continue

        locals()[v1_method.__name__] = functools.partialmethod(callback_impl, name=v1_method.__name__)

    def playbook_on_stats(self, stats, *args, **kwargs):
        if missed_v1_method_calls := (
                {'on_any',
                 'runner_on_ok',
                 'playbook_on_task_start',
                 'runner_on_async_ok',
                 } - self.called_v1_method_names):
            assert False, f"The following v1 callback methods were not invoked as expected: {', '.join(missed_v1_method_calls)}"

        print("v1 callback test PASS")



# (c) 2012, Jeroen Hoekx <jeroen@hoekx.be>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import base64
import functools
import glob
import hashlib
import json
import ntpath
import os.path
import re
import shlex
import time
import uuid
import yaml
import datetime
import typing as t

from collections.abc import Mapping
from functools import partial
from random import Random, SystemRandom, shuffle

from jinja2.filters import do_map, do_select, do_selectattr, do_reject, do_rejectattr, pass_environment, sync_do_groupby
from jinja2.environment import Environment

from ansible._internal._templating import _lazy_containers
from ansible.errors import AnsibleFilterError, AnsibleTypeError, AnsibleTemplatePluginError
from ansible.module_utils.datatag import native_type_name
from ansible.module_utils.common.json import get_encoder, get_decoder
from ansible.module_utils.six import string_types, integer_types, text_type
from ansible.module_utils.common.text.converters import to_bytes, to_native, to_text
from ansible.module_utils.common.collections import is_sequence
from ansible.parsing.yaml.dumper import AnsibleDumper
from ansible.template import accept_args_markers, accept_lazy_markers
from ansible._internal._templating._jinja_common import MarkerError, UndefinedMarker, validate_arg_type
from ansible._internal._yaml import _loader as _yaml_loader
from ansible.utils.display import Display
from ansible.utils.encrypt import do_encrypt, PASSLIB_AVAILABLE
from ansible.utils.hashing import md5s, checksum_s
from ansible.utils.unicode import unicode_wrap
from ansible.utils.vars import merge_hash

display = Display()

UUID_NAMESPACE_ANSIBLE = uuid.UUID('361E6D51-FAEC-444A-9079-341386DA8E2E')


@accept_lazy_markers
def to_yaml(a, *_args, default_flow_style: bool | None = None, **kwargs) -> str:
    """Serialize input as terse flow-style YAML."""
    return yaml.dump(a, Dumper=AnsibleDumper, allow_unicode=True, default_flow_style=default_flow_style, **kwargs)


@accept_lazy_markers
def to_nice_yaml(a, indent=4, *_args, default_flow_style=False, **kwargs) -> str:
    """Serialize input as verbose multi-line YAML."""
    return to_yaml(a, indent=indent, default_flow_style=default_flow_style, **kwargs)


def from_json(a, profile: str | None = None, **kwargs) -> t.Any:
    """Deserialize JSON with an optional decoder profile."""
    cls = get_decoder(profile or "tagless")

    return json.loads(a, cls=cls, **kwargs)


def to_json(a, profile: str | None = None, vault_to_text: t.Any = ..., preprocess_unsafe: t.Any = ..., **kwargs) -> str:
    """Serialize as JSON with an optional encoder profile."""

    if profile and vault_to_text is not ...:
        raise ValueError("Only one of `vault_to_text` or `profile` can be specified.")

    if profile and preprocess_unsafe is not ...:
        raise ValueError("Only one of `preprocess_unsafe` or `profile` can be specified.")

    # deprecated: description='deprecate vault_to_text' core_version='2.23'
    # deprecated: description='deprecate preprocess_unsafe' core_version='2.23'

    cls = get_encoder(profile or "tagless")

    return json.dumps(a, cls=cls, **kwargs)


def to_nice_json(a, indent=4, sort_keys=True, **kwargs):
    """Make verbose, human-readable JSON."""
    # TODO separators can be potentially exposed to the user as well
    kwargs.pop('separators', None)
    return to_json(a, indent=indent, sort_keys=sort_keys, separators=(',', ': '), **kwargs)


# CAUTION: Do not put non-string values here since they can have unwanted logical equality, such as 1.0 (equal to 1 and True) or 0.0 (equal to 0 and False).
_valid_bool_true = {'yes', 'on', 'true', '1'}
_valid_bool_false = {'no', 'off', 'false', '0'}


def to_bool(value: object) -> bool:
    """Convert well-known input values to a boolean value."""
    value_to_check: object

    if isinstance(value, str):
        value_to_check = value.lower()  # accept mixed case variants
    elif isinstance(value, int):  # bool is also an int
        value_to_check = str(value).lower()  # accept int (0, 1) and bool (True, False) -- not just string versions
    else:
        value_to_check = value

    try:
        if value_to_check in _valid_bool_true:
            return True

        if value_to_check in _valid_bool_false:
            return False

        # if we're still here, the value is unsupported- always fire a deprecation warning
        result = value_to_check == 1  # backwards compatibility with the old code which checked: value in ('yes', 'on', '1', 'true', 1)
    except TypeError:
        result = False

    # NB: update the doc string to reflect reality once this fallback is removed
    display.deprecated(
        msg=f'The `bool` filter coerced invalid value {value!r} ({native_type_name(value)}) to {result!r}.',
        version='2.23',
    )

    return result


def to_datetime(string, format="%Y-%m-%d %H:%M:%S"):
    return datetime.datetime.strptime(string, format)


def strftime(string_format, second=None, utc=False):
    """ return a date string using string. See https://docs.python.org/3/library/time.html#time.strftime for format """
    if utc:
        timefn = time.gmtime
    else:
        timefn = time.localtime
    if second is not None:
        try:
            second = float(second)
        except Exception:
            raise AnsibleFilterError('Invalid value for epoch value (%s)' % second)
    return time.strftime(string_format, timefn(second))


def quote(a):
    """ return its argument quoted for shell usage """
    if a is None:
        a = u''
    return shlex.quote(to_text(a))


def fileglob(pathname):
    """ return list of matched regular files for glob """
    return [g for g in glob.glob(pathname) if os.path.isfile(g)]


def regex_replace(value='', pattern='', replacement='', ignorecase=False, multiline=False, count=0, mandatory_count=0):
    """ Perform a `re.sub` returning a string """

    value = to_text(value, errors='surrogate_or_strict', nonstring='simplerepr')

    flags = 0
    if ignorecase:
        flags |= re.I
    if multiline:
        flags |= re.M
    _re = re.compile(pattern, flags=flags)
    (output, subs) = _re.subn(replacement, value, count=count)
    if mandatory_count and mandatory_count != subs:
        raise AnsibleFilterError("'%s' should match %d times, but matches %d times in '%s'"
                                 % (pattern, mandatory_count, count, value))
    return output


def regex_findall(value, regex, multiline=False, ignorecase=False):
    """ Perform re.findall and return the list of matches """

    value = to_text(value, errors='surrogate_or_strict', nonstring='simplerepr')

    flags = 0
    if ignorecase:
        flags |= re.I
    if multiline:
        flags |= re.M
    return re.findall(regex, value, flags)


def regex_search(value, regex, *args, **kwargs):
    """ Perform re.search and return the list of matches or a backref """

    value = to_text(value, errors='surrogate_or_strict', nonstring='simplerepr')

    groups = list()
    for arg in args:
        if arg.startswith('\\g'):
            match = re.match(r'\\g<(\S+)>', arg).group(1)
            groups.append(match)
        elif arg.startswith('\\'):
            match = int(re.match(r'\\(\d+)', arg).group(1))
            groups.append(match)
        else:
            raise AnsibleFilterError('Unknown argument')

    flags = 0
    if kwargs.get('ignorecase'):
        flags |= re.I
    if kwargs.get('multiline'):
        flags |= re.M

    match = re.search(regex, value, flags)
    if match:
        if not groups:
            return match.group()
        else:
            items = list()
            for item in groups:
                items.append(match.group(item))
            return items


@accept_args_markers
def ternary(value, true_val, false_val, none_val=None):
    """  value ? true_val : false_val """
    if value is None and none_val is not None:
        return none_val
    elif bool(value):
        return true_val
    else:
        return false_val


def regex_escape(string, re_type='python'):
    """Escape all regular expressions special characters from STRING."""
    string = to_text(string, errors='surrogate_or_strict', nonstring='simplerepr')
    if re_type == 'python':
        return re.escape(string)
    elif re_type == 'posix_basic':
        # list of BRE special chars:
        # https://en.wikibooks.org/wiki/Regular_Expressions/POSIX_Basic_Regular_Expressions
        return regex_replace(string, r'([].[^$*\\])', r'\\\1')
    # TODO: implement posix_extended
    # It's similar to, but different from python regex, which is similar to,
    # but different from PCRE.  It's possible that re.escape would work here.
    # https://remram44.github.io/regex-cheatsheet/regex.html#programs
    elif re_type == 'posix_extended':
        raise AnsibleFilterError('Regex type (%s) not yet implemented' % re_type)
    else:
        raise AnsibleFilterError('Invalid regex type (%s)' % re_type)


def from_yaml(data):
    if data is None:
        return None

    if isinstance(data, str):
        return yaml.load(data, Loader=_yaml_loader.AnsibleInstrumentedLoader)  # type: ignore[arg-type]

    display.deprecated(f"The from_yaml filter ignored non-string input of type {native_type_name(data)!r}.", version='2.23', obj=data)
    return data


def from_yaml_all(data):
    if data is None:
        return []  # backward compatibility; ensure consistent result between classic/native Jinja for None/empty string input

    if isinstance(data, str):
        return yaml.load_all(data, Loader=_yaml_loader.AnsibleInstrumentedLoader)  # type: ignore[arg-type]

    display.deprecated(f"The from_yaml_all filter ignored non-string input of type {native_type_name(data)!r}.", version='2.23', obj=data)
    return data


@pass_environment
def rand(environment, end, start=None, step=None, seed=None):
    if seed is None:
        r = SystemRandom()
    else:
        r = Random(seed)
    if isinstance(end, integer_types):
        if not start:
            start = 0
        if not step:
            step = 1
        return r.randrange(start, end, step)
    elif hasattr(end, '__iter__'):
        if start or step:
            raise AnsibleFilterError('start and step can only be used with integer values')
        return r.choice(end)
    else:
        raise AnsibleFilterError('random can only be used on sequences and integers')


def randomize_list(mylist, seed=None):
    try:
        mylist = list(mylist)
        if seed:
            r = Random(seed)
            r.shuffle(mylist)
        else:
            shuffle(mylist)
    except Exception:
        pass
    return mylist


def get_hash(data, hashtype='sha1'):
    try:
        h = hashlib.new(hashtype)
    except Exception as e:
        # hash is not supported?
        raise AnsibleFilterError(e)

    h.update(to_bytes(data, errors='surrogate_or_strict'))
    return h.hexdigest()


def get_encrypted_password(password, hashtype='sha512', salt=None, salt_size=None, rounds=None, ident=None):
    passlib_mapping = {
        'md5': 'md5_crypt',
        'blowfish': 'bcrypt',
        'sha256': 'sha256_crypt',
        'sha512': 'sha512_crypt',
    }

    hashtype = passlib_mapping.get(hashtype, hashtype)

    if PASSLIB_AVAILABLE and hashtype not in passlib_mapping and hashtype not in passlib_mapping.values():
        raise AnsibleFilterError(f"{hashtype} is not in the list of supported passlib algorithms: {', '.join(passlib_mapping)}")

    return do_encrypt(password, hashtype, salt=salt, salt_size=salt_size, rounds=rounds, ident=ident)


def to_uuid(string, namespace=UUID_NAMESPACE_ANSIBLE):
    uuid_namespace = namespace
    if not isinstance(uuid_namespace, uuid.UUID):
        try:
            uuid_namespace = uuid.UUID(namespace)
        except (AttributeError, ValueError) as e:
            raise AnsibleFilterError("Invalid value '%s' for 'namespace': %s" % (to_native(namespace), to_native(e)))
    # uuid.uuid5() requires bytes on Python 2 and bytes or text or Python 3
    return to_text(uuid.uuid5(uuid_namespace, to_native(string, errors='surrogate_or_strict')))


@accept_args_markers
def mandatory(a: object, msg: str | None = None) -> object:
    """Make a variable mandatory."""
    # DTFIX-FUTURE: deprecate this filter; there are much better ways via undef, etc...
    #                also remember to remove unit test checking for _undefined_name
    if isinstance(a, UndefinedMarker):
        if msg is not None:
            raise AnsibleFilterError(to_text(msg))

        if a._undefined_name is not None:
            name = f'{to_text(a._undefined_name)!r} '
        else:
            name = ''

        raise AnsibleFilterError(f"Mandatory variable {name}not defined.")

    return a


def combine(*terms, **kwargs):
    recursive = kwargs.pop('recursive', False)
    list_merge = kwargs.pop('list_merge', 'replace')
    if kwargs:
        raise AnsibleFilterError("'recursive' and 'list_merge' are the only valid keyword arguments")

    # allow the user to do `[dict1, dict2, ...] | combine`
    dictionaries = flatten(terms, levels=1)

    if not dictionaries:
        return {}

    if len(dictionaries) == 1:
        return dictionaries[0]

    # merge all the dicts so that the dict at the end of the array have precedence
    # over the dict at the beginning.
    # we merge the dicts from the highest to the lowest priority because there is
    # a huge probability that the lowest priority dict will be the biggest in size
    # (as the low prio dict will hold the "default" values and the others will be "patches")
    # and merge_hash create a copy of it's first argument.
    # so high/right -> low/left is more efficient than low/left -> high/right
    high_to_low_prio_dict_iterator = reversed(dictionaries)
    result = next(high_to_low_prio_dict_iterator)
    for dictionary in high_to_low_prio_dict_iterator:
        result = merge_hash(dictionary, result, recursive, list_merge)

    return result


def comment(text, style='plain', **kw):
    # Predefined comment types
    comment_styles = {
        'plain': {
            'decoration': '# '
        },
        'erlang': {
            'decoration': '% '
        },
        'c': {
            'decoration': '// '
        },
        'cblock': {
            'beginning': '/*',
            'decoration': ' * ',
            'end': ' */'
        },
        'xml': {
            'beginning': '<!--',
            'decoration': ' - ',
            'end': '-->'
        }
    }

    if style not in comment_styles:
        raise AnsibleTemplatePluginError(
            message=f"Invalid style {style!r}.",
            help_text=f"Available styles: {', '.join(comment_styles)}",
            obj=style,
        )

    # Pointer to the right comment type
    style_params = comment_styles[style]

    if 'decoration' in kw:
        prepostfix = kw['decoration']
    else:
        prepostfix = style_params['decoration']

    # Default params
    p = {
        'newline': '\n',
        'beginning': '',
        'prefix': (prepostfix).rstrip(),
        'prefix_count': 1,
        'decoration': '',
        'postfix': (prepostfix).rstrip(),
        'postfix_count': 1,
        'end': ''
    }

    # Update default params
    p.update(style_params)
    p.update(kw)

    # Compose substrings for the final string
    str_beginning = ''
    if p['beginning']:
        str_beginning = "%s%s" % (p['beginning'], p['newline'])
    str_prefix = ''
    if p['prefix']:
        if p['prefix'] != p['newline']:
            str_prefix = str(
                "%s%s" % (p['prefix'], p['newline'])) * int(p['prefix_count'])
        else:
            str_prefix = str(
                "%s" % (p['newline'])) * int(p['prefix_count'])
    str_text = ("%s%s" % (
        p['decoration'],
        # Prepend each line of the text with the decorator
        text.replace(
            p['newline'], "%s%s" % (p['newline'], p['decoration'])))).replace(
                # Remove trailing spaces when only decorator is on the line
                "%s%s" % (p['decoration'], p['newline']),
                "%s%s" % (p['decoration'].rstrip(), p['newline']))
    str_postfix = p['newline'].join(
        [''] + [p['postfix'] for x in range(p['postfix_count'])])
    str_end = ''
    if p['end']:
        str_end = "%s%s" % (p['newline'], p['end'])

    # Return the final string
    return "%s%s%s%s%s" % (
        str_beginning,
        str_prefix,
        str_text,
        str_postfix,
        str_end)


@pass_environment
def extract(environment: Environment, item, container, morekeys=None):
    if morekeys is None:
        keys = [item]
    elif isinstance(morekeys, list):
        keys = [item] + morekeys
    else:
        keys = [item, morekeys]

    value = container

    for key in keys:
        try:
            value = environment.getitem(value, key)
        except MarkerError as ex:
            value = ex.source

    return value


def b64encode(string, encoding='utf-8', urlsafe=False):
    func = base64.b64encode
    if urlsafe:
        func = base64.urlsafe_b64encode
    return to_text(func(to_bytes(string, encoding=encoding, errors='surrogate_or_strict')))


def b64decode(string, encoding='utf-8', urlsafe=False):
    func = base64.b64decode
    if urlsafe:
        func = base64.urlsafe_b64decode
    return to_text(func(to_bytes(string, errors='surrogate_or_strict')), encoding=encoding)


def flatten(mylist, levels=None, skip_nulls=True):

    ret = []
    for element in mylist:
        if skip_nulls and element in (None, 'None', 'null'):
            # ignore null items
            continue
        elif is_sequence(element):
            if levels is None:
                ret.extend(flatten(element, skip_nulls=skip_nulls))
            elif levels >= 1:
                # decrement as we go down the stack
                ret.extend(flatten(element, levels=(int(levels) - 1), skip_nulls=skip_nulls))
            else:
                ret.append(element)
        else:
            ret.append(element)

    return ret


def subelements(obj, subelements, skip_missing=False):
    """Accepts a dict or list of dicts, and a dotted accessor and produces a product
    of the element and the results of the dotted accessor

    >>> obj = [{"name": "alice", "groups": ["wheel"], "authorized": ["/tmp/alice/onekey.pub"]}]
    >>> subelements(obj, 'groups')
    [({'name': 'alice', 'groups': ['wheel'], 'authorized': ['/tmp/alice/onekey.pub']}, 'wheel')]

    """
    if isinstance(obj, dict):
        element_list = list(obj.values())
    elif isinstance(obj, list):
        element_list = obj[:]
    else:
        raise AnsibleFilterError('obj must be a list of dicts or a nested dict')

    if isinstance(subelements, list):
        subelement_list = subelements[:]
    elif isinstance(subelements, string_types):
        subelement_list = subelements.split('.')
    else:
        raise AnsibleTypeError('subelements must be a list or a string')

    results = []

    for element in element_list:
        values = element
        for subelement in subelement_list:
            try:
                values = values[subelement]
            except KeyError:
                if skip_missing:
                    values = []
                    break
                raise AnsibleFilterError("could not find %r key in iterated item %r" % (subelement, values))
            except TypeError as ex:
                raise AnsibleTypeError("the key %s should point to a dictionary, got '%s'" % (subelement, values)) from ex
        if not isinstance(values, list):
            raise AnsibleTypeError("the key %r should point to a list, got %r" % (subelement, values))

        for value in values:
            results.append((element, value))

    return results


def dict_to_list_of_dict_key_value_elements(mydict, key_name='key', value_name='value'):
    """ takes a dictionary and transforms it into a list of dictionaries,
        with each having a 'key' and 'value' keys that correspond to the keys and values of the original """

    if not isinstance(mydict, Mapping):
        raise AnsibleTypeError("dict2items requires a dictionary, got %s instead." % type(mydict))

    ret = []
    for key in mydict:
        ret.append({key_name: key, value_name: mydict[key]})
    return ret


def list_of_dict_key_value_elements_to_dict(mylist, key_name='key', value_name='value'):
    """ takes a list of dicts with each having a 'key' and 'value' keys, and transforms the list into a dictionary,
        effectively as the reverse of dict2items """

    if not is_sequence(mylist):
        raise AnsibleTypeError("items2dict requires a list, got %s instead." % type(mylist))

    try:
        return dict((item[key_name], item[value_name]) for item in mylist)
    except KeyError:
        raise AnsibleTypeError(
            "items2dict requires each dictionary in the list to contain the keys '%s' and '%s', got %s instead."
            % (key_name, value_name, mylist)
        )
    except TypeError:
        raise AnsibleTypeError("items2dict requires a list of dictionaries, got %s instead." % mylist)


def path_join(paths):
    """ takes a sequence or a string, and return a concatenation
        of the different members """
    if isinstance(paths, string_types):
        return os.path.join(paths)
    if is_sequence(paths):
        return os.path.join(*paths)
    raise AnsibleTypeError("|path_join expects string or sequence, got %s instead." % type(paths))


def commonpath(paths):
    """
    Retrieve the longest common path from the given list.

    :param paths: A list of file system paths.
    :type paths: List[str]
    :returns: The longest common path.
    :rtype: str
    """
    if not is_sequence(paths):
        raise AnsibleTypeError("|commonpath expects sequence, got %s instead." % type(paths))

    return os.path.commonpath(paths)


class GroupTuple(t.NamedTuple):
    """
    Custom named tuple for the groupby filter with a public interface; silently ignored by unknown type checks.
    This matches the internal implementation of the _GroupTuple returned by Jinja's built-in groupby filter.
    """

    grouper: t.Any
    list: list[t.Any]

    def __repr__(self) -> str:
        return tuple.__repr__(self)


_lazy_containers.register_known_types(GroupTuple)


@pass_environment
def _cleansed_groupby(*args, **kwargs):
    res = sync_do_groupby(*args, **kwargs)
    res = [GroupTuple(grouper=g.grouper, list=g.list) for g in res]

    return res

# DTFIX-FUTURE: make these dumb wrappers more dynamic


@accept_args_markers
def ansible_default(
    value: t.Any,
    default_value: t.Any = '',
    boolean: bool = False,
) -> t.Any:
    """Updated `default` filter that only coalesces classic undefined objects; other Undefined-derived types (eg, ErrorMarker) pass through."""
    validate_arg_type('boolean', boolean, bool)

    if isinstance(value, UndefinedMarker):
        return default_value

    if boolean and not value:
        return default_value

    return value


@accept_lazy_markers
@functools.wraps(do_map)
def wrapped_map(*args, **kwargs) -> t.Any:
    return do_map(*args, **kwargs)


@accept_lazy_markers
@functools.wraps(do_select)
def wrapped_select(*args, **kwargs) -> t.Any:
    return do_select(*args, **kwargs)


@accept_lazy_markers
@functools.wraps(do_selectattr)
def wrapped_selectattr(*args, **kwargs) -> t.Any:
    return do_selectattr(*args, **kwargs)


@accept_lazy_markers
@functools.wraps(do_reject)
def wrapped_reject(*args, **kwargs) -> t.Any:
    return do_reject(*args, **kwargs)


@accept_lazy_markers
@functools.wraps(do_rejectattr)
def wrapped_rejectattr(*args, **kwargs) -> t.Any:
    return do_rejectattr(*args, **kwargs)


@accept_args_markers
def type_debug(obj: object) -> str:
    return native_type_name(obj)


class FilterModule(object):
    """ Ansible core jinja2 filters """

    def filters(self):
        return {
            # base 64
            'b64decode': b64decode,
            'b64encode': b64encode,

            # uuid
            'to_uuid': to_uuid,

            # json
            'to_json': to_json,
            'to_nice_json': to_nice_json,
            'from_json': from_json,

            # yaml
            'to_yaml': to_yaml,
            'to_nice_yaml': to_nice_yaml,
            'from_yaml': from_yaml,
            'from_yaml_all': from_yaml_all,

            # path
            'basename': partial(unicode_wrap, os.path.basename),
            'dirname': partial(unicode_wrap, os.path.dirname),
            'expanduser': partial(unicode_wrap, os.path.expanduser),
            'expandvars': partial(unicode_wrap, os.path.expandvars),
            'path_join': path_join,
            'realpath': partial(unicode_wrap, os.path.realpath),
            'relpath': partial(unicode_wrap, os.path.relpath),
            'splitext': partial(unicode_wrap, os.path.splitext),
            'win_basename': partial(unicode_wrap, ntpath.basename),
            'win_dirname': partial(unicode_wrap, ntpath.dirname),
            'win_splitdrive': partial(unicode_wrap, ntpath.splitdrive),
            'commonpath': commonpath,
            'normpath': partial(unicode_wrap, os.path.normpath),

            # file glob
            'fileglob': fileglob,

            # types
            'bool': to_bool,
            'to_datetime': to_datetime,

            # date formatting
            'strftime': strftime,

            # quote string for shell usage
            'quote': quote,

            # hash filters
            # md5 hex digest of string
            'md5': md5s,
            # sha1 hex digest of string
            'sha1': checksum_s,
            # checksum of string as used by ansible for checksumming files
            'checksum': checksum_s,
            # generic hashing
            'password_hash': get_encrypted_password,
            'hash': get_hash,

            # regex
            'regex_replace': regex_replace,
            'regex_escape': regex_escape,
            'regex_search': regex_search,
            'regex_findall': regex_findall,

            # ? : ;
            'ternary': ternary,

            # random stuff
            'random': rand,
            'shuffle': randomize_list,

            # undefined
            'mandatory': mandatory,

            # comment-style decoration
            'comment': comment,

            # debug
            'type_debug': type_debug,

            # Data structures
            'combine': combine,
            'extract': extract,
            'flatten': flatten,
            'dict2items': dict_to_list_of_dict_key_value_elements,
            'items2dict': list_of_dict_key_value_elements_to_dict,
            'subelements': subelements,
            'split': partial(unicode_wrap, text_type.split),
            # FDI038 - replace this with a standard type compat shim
            'groupby': _cleansed_groupby,

            # Jinja builtins that need special arg handling
            'd': ansible_default,  # replaces the implementation instead of wrapping it
            'default': ansible_default,  # replaces the implementation instead of wrapping it
            'map': wrapped_map,
            'select': wrapped_select,
            'selectattr': wrapped_selectattr,
            'reject': wrapped_reject,
            'rejectattr': wrapped_rejectattr,
        }



# (c) 2012, Jan-Piet Mens <jpmens(at)gmail.com>
# (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

DOCUMENTATION = """
    name: env
    author: Jan-Piet Mens (@jpmens) <jpmens(at)gmail.com>
    version_added: "0.9"
    short_description: Read the value of environment variables
    description:
      - Allows you to query the environment variables available on the
        controller when you invoked Ansible.
    options:
      _terms:
        description:
          - Environment variable or list of them to lookup the values for.
        required: True
      default:
        description: What return when the variable is undefined
        type: raw
        default: ''
        version_added: '2.13'
    notes:
        - You can pass the C(Undefined) object as O(default) to force an undefined error
"""

EXAMPLES = """
- name: Basic usage
  ansible.builtin.debug:
    msg: "'{{ lookup('ansible.builtin.env', 'HOME') }}' is the HOME environment variable."

- name: Before 2.13, how to set default value if the variable is not defined
  ansible.builtin.debug:
    msg: "Hello {{ lookup('ansible.builtin.env', 'UNDEFINED_VARIABLE') | default('World', True) }}"

- name: Example how to set default value if the variable is not defined
  ansible.builtin.debug:
    msg: "Hello {{ lookup('ansible.builtin.env', 'UNDEFINED_VARIABLE', default='World') }}"

- name: Fail if the variable is not defined by setting default value to 'undef()'
  ansible.builtin.debug:
    msg: "Hello {{ lookup('ansible.builtin.env', 'UNDEFINED_VARIABLE', default=undef()) }}"
"""

RETURN = """
  _list:
    description:
      - Values from the environment variables.
    type: list
"""

import os


from ansible.plugins.lookup import LookupBase
from ansible._internal._templating._jinja_bits import _undef, _DEFAULT_UNDEF
from ansible._internal._datatag._tags import Origin


class LookupModule(LookupBase):
    accept_args_markers = True  # the `default` arg can accept undefined values

    def run(self, terms, variables=None, **kwargs):
        self.set_options(var_options=variables, direct=kwargs)

        ret = []
        d = self.get_option('default')

        for term in terms:
            var = term.split()[0]
            val = os.environ.get(var, d)

            if val is _DEFAULT_UNDEF:
                val = _undef(f'The environment variable {var!r} is not set.')
            else:
                val = Origin(description=f"<environment variable {var!r}>").try_tag(val)

            ret.append(val)

        return ret



# (c) 2016 RedHat
#
# This file is part of Ansible.
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import dataclasses
import os
import os.path
import re
import secrets
import shlex
import time

from collections.abc import Mapping, Sequence

from ansible.errors import AnsibleError
from ansible.module_utils.common.text.converters import to_native
from ansible.module_utils.six import text_type, string_types
from ansible.plugins import AnsiblePlugin

_USER_HOME_PATH_RE = re.compile(r'^~[_.A-Za-z0-9][-_.A-Za-z0-9]*$')


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class _ShellCommand:
    """Internal type returned by shell subsystems that may require both an execution payload and a command (eg powershell)."""
    command: str
    input_data: bytes | None = None


class ShellBase(AnsiblePlugin):
    def __init__(self):

        super(ShellBase, self).__init__()

        # Not used but here for backwards compatibility.
        # ansible.posix.fish uses (but does not actually use) this value.
        # https://github.com/ansible-collections/ansible.posix/blob/f41f08e9e3d3129e709e122540b5ae6bc19932be/plugins/shell/fish.py#L38-L39
        self.env = {}
        self.tmpdir = None
        self.executable = None

    def _normalize_system_tmpdirs(self):
        # Normalize the tmp directory strings. We don't use expanduser/expandvars because those
        # can vary between remote user and become user.  Therefore the safest practice will be for
        # this to always be specified as full paths)
        normalized_paths = [d.rstrip('/') for d in self.get_option('system_tmpdirs')]

        # Make sure all system_tmpdirs are absolute otherwise they'd be relative to the login dir
        # which is almost certainly going to fail in a cornercase.
        if not all(os.path.isabs(d) for d in normalized_paths):
            raise AnsibleError('The configured system_tmpdirs contains a relative path: {0}. All'
                               ' system_tmpdirs must be absolute'.format(to_native(normalized_paths)))

        self.set_option('system_tmpdirs', normalized_paths)

    def set_options(self, task_keys=None, var_options=None, direct=None):

        super(ShellBase, self).set_options(task_keys=task_keys, var_options=var_options, direct=direct)

        # We can remove the try: except in the future when we make ShellBase a proper subset of
        # *all* shells.  Right now powershell and third party shells which do not use the
        # shell_common documentation fragment (and so do not have system_tmpdirs) will fail
        try:
            self._normalize_system_tmpdirs()
        except KeyError:
            pass

    @staticmethod
    def _generate_temp_dir_name():
        return 'ansible-tmp-%s-%s-%s' % (time.time(), os.getpid(), secrets.randbelow(2**48))

    def env_prefix(self, **kwargs):
        return ' '.join(['%s=%s' % (k, self.quote(text_type(v))) for k, v in kwargs.items()])

    def join_path(self, *args):
        return os.path.join(*args)

    # some shells (eg, powershell) are snooty about filenames/extensions, this lets the shell plugin have a say
    def get_remote_filename(self, pathname):
        base_name = os.path.basename(pathname.strip())
        return base_name.strip()

    def path_has_trailing_slash(self, path):
        return path.endswith('/')

    def chmod(self, paths, mode):
        cmd = ['chmod', mode]
        cmd.extend(paths)
        return self.join(cmd)

    def chown(self, paths, user):
        cmd = ['chown', user]
        cmd.extend(paths)
        return self.join(cmd)

    def chgrp(self, paths, group):
        cmd = ['chgrp', group]
        cmd.extend(paths)
        return self.join(cmd)

    def set_user_facl(self, paths, user, mode):
        """Only sets acls for users as that's really all we need"""
        cmd = ['setfacl', '-m', 'u:%s:%s' % (user, mode)]
        cmd.extend(paths)
        return self.join(cmd)

    def remove(self, path, recurse=False):
        path = self.quote(path)
        cmd = 'rm -f '
        if recurse:
            cmd += '-r '
        return cmd + "%s %s" % (path, self._SHELL_REDIRECT_ALLNULL)

    def exists(self, path):
        cmd = ['test', '-e', self.quote(path)]
        return ' '.join(cmd)

    def mkdtemp(
        self,
        basefile: str | None = None,
        system: bool = False,
        mode: int = 0o700,
        tmpdir: str | None = None,
    ) -> str:
        if not basefile:
            basefile = self.__class__._generate_temp_dir_name()

        # When system is specified we have to create this in a directory where
        # other users can read and access the tmp directory.
        # This is because we use system to create tmp dirs for unprivileged users who are
        # sudo'ing to a second unprivileged user.
        # The 'system_tmpdirs' setting defines directories we can use for this purpose
        # the default are, /tmp and /var/tmp.
        # So we only allow one of those locations if system=True, using the
        # passed in tmpdir if it is valid or the first one from the setting if not.

        if system:
            if tmpdir:
                tmpdir = tmpdir.rstrip('/')

            if tmpdir in self.get_option('system_tmpdirs'):
                basetmpdir = tmpdir
            else:
                basetmpdir = self.get_option('system_tmpdirs')[0]
        else:
            if tmpdir is None:
                basetmpdir = self.get_option('remote_tmp')
            else:
                basetmpdir = tmpdir

        basetmp = self.join_path(basetmpdir, basefile)

        # use mkdir -p to ensure parents exist, but mkdir fullpath to ensure last one is created by us
        cmd = 'mkdir -p %s echo %s %s' % (self._SHELL_SUB_LEFT, basetmpdir, self._SHELL_SUB_RIGHT)
        cmd += '%s mkdir %s echo %s %s' % (self._SHELL_AND, self._SHELL_SUB_LEFT, basetmp, self._SHELL_SUB_RIGHT)
        cmd += ' %s echo %s=%s echo %s %s' % (self._SHELL_AND, basefile, self._SHELL_SUB_LEFT, basetmp, self._SHELL_SUB_RIGHT)

        # change the umask in a subshell to achieve the desired mode
        # also for directories created with `mkdir -p`
        if mode:
            tmp_umask = 0o777 & ~mode
            cmd = '%s umask %o %s %s %s' % (self._SHELL_GROUP_LEFT, tmp_umask, self._SHELL_AND, cmd, self._SHELL_GROUP_RIGHT)

        return cmd

    def _mkdtemp2(
        self,
        basefile: str | None = None,
        system: bool = False,
        mode: int = 0o700,
        tmpdir: str | None = None,
    ) -> _ShellCommand:
        """Gets command info to create a temporary directory.

        This is an internal API that should not be used publicly.

        :args basefile: The base name of the temporary directory.
        :args system: If True, create the directory in a system-wide location.
        :args mode: The permissions mode for the directory.
        :args tmpdir: The directory in which to create the temporary directory.
        :returns: The shell command to run to create the temp directory.
        """
        cmd = self.mkdtemp(basefile=basefile, system=system, mode=mode, tmpdir=tmpdir)
        return _ShellCommand(command=cmd, input_data=None)

    def expand_user(
        self,
        user_home_path: str,
        username: str = '',
    ) -> str:
        """ Return a command to expand tildes in a path

        It can be either "~" or "~username". We just ignore $HOME
        We use the POSIX definition of a username:
            http://pubs.opengroup.org/onlinepubs/000095399/basedefs/xbd_chap03.html#tag_03_426
            http://pubs.opengroup.org/onlinepubs/000095399/basedefs/xbd_chap03.html#tag_03_276

            Falls back to 'current working directory' as we assume 'home is where the remote user ends up'
        """

        # Check that the user_path to expand is safe
        if user_home_path != '~':
            if not _USER_HOME_PATH_RE.match(user_home_path):
                user_home_path = self.quote(user_home_path)
        elif username:
            # if present the user name is appended to resolve "that user's home"
            user_home_path += username

        return 'echo %s' % user_home_path

    def _expand_user2(
        self,
        user_home_path: str,
        username: str = '',
    ) -> _ShellCommand:
        """Gets command to expand user path.

        This is an internal API that should not be used publicly.

        :args user_home_path: The path to expand.
        :args username: The username to use for expansion.
        :returns: The shell command to run to get the expanded user path.
        """
        cmd = self.expand_user(user_home_path, username=username)
        return _ShellCommand(command=cmd, input_data=None)

    def pwd(self):
        """Return the working directory after connecting"""
        return 'echo %spwd%s' % (self._SHELL_SUB_LEFT, self._SHELL_SUB_RIGHT)

    def build_module_command(self, env_string, shebang, cmd, arg_path=None):
        env_string = env_string.strip()
        if env_string:
            env_string += ' '

        if shebang is None:
            shebang = ''

        cmd_parts = [
            shebang.removeprefix('#!').strip(),
            cmd.strip(),
            arg_path,
        ]

        cleaned_up_cmd = self.join(
            stripped_cmd_part for raw_cmd_part in cmd_parts
            if raw_cmd_part and (stripped_cmd_part := raw_cmd_part.strip())
        )
        return ''.join((env_string, cleaned_up_cmd))

    def append_command(self, cmd, cmd_to_append):
        """Append an additional command if supported by the shell"""

        if self._SHELL_AND:
            cmd += ' %s %s' % (self._SHELL_AND, cmd_to_append)

        return cmd

    def wrap_for_exec(self, cmd):
        """wrap script execution with any necessary decoration (eg '&' for quoted powershell script paths)"""
        return cmd

    def quote(self, cmd):
        """Returns a shell-escaped string that can be safely used as one token in a shell command line"""
        return shlex.quote(cmd)

    def join(self, cmd_parts):
        """Returns a shell-escaped string from a list that can be safely used in a shell command line"""
        return shlex.join(cmd_parts)



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
# (c) 2017 Ansible Project
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import pytest

from ansible.cli.arguments import option_helpers as opt_help
from ansible.utils import context_objects as co


@pytest.fixture
def parser():
    parser = opt_help.create_base_parser('testparser')

    opt_help.add_runas_options(parser)
    opt_help.add_meta_options(parser)
    opt_help.add_runtask_options(parser)
    opt_help.add_vault_options(parser)
    opt_help.add_async_options(parser)
    opt_help.add_connect_options(parser)
    opt_help.add_subset_options(parser)
    opt_help.add_check_options(parser)
    opt_help.add_inventory_options(parser)

    return parser


@pytest.fixture
def reset_cli_args():
    co.GlobalCLIArgs._Singleton__instance = None
    yield
    co.GlobalCLIArgs._Singleton__instance = None



"""Temporary plugin to prevent stdout noise pollution from finalization of abandoned generators."""

from __future__ import annotations

import sys
import typing as t

if t.TYPE_CHECKING:
    from pylint.lint import PyLinter


def _mask_finalizer_valueerror(ur: t.Any) -> None:
    """Mask only ValueErrors from finalizing abandoned generators; delegate everything else"""
    # work around Python finalizer issue that sometimes spews this error message to stdout
    # see https://github.com/pylint-dev/pylint/issues/9138
    if ur.exc_type is ValueError and 'generator already executing' in str(ur.exc_value):
        return

    sys.__unraisablehook__(ur)


def register(linter: PyLinter) -> None:  # pylint: disable=unused-argument
    """PyLint plugin registration entrypoint"""
    sys.unraisablehook = _mask_finalizer_valueerror



#!/usr/bin/python

from __future__ import annotations

import tempfile

from dataclasses import dataclass

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.respawn import has_respawned, probe_interpreters_for_module, respawn_module

HAS_RPMFLUFF = True

try:
    from rpmfluff.make import make_gif
    from rpmfluff.sourcefile import GeneratedSourceFile
    from rpmfluff.rpmbuild import SimpleRpmBuild
    from rpmfluff.utils import expectedArch
    from rpmfluff.yumrepobuild import YumRepoBuild
except ImportError:
    expectedArch = None  # define here to avoid NameError as it is used on top level in SPECS
    HAS_RPMFLUFF = False


@dataclass
class RPM:
    name: str
    version: str
    release: str = '1'
    epoch: int = 0
    arch: list[str] | None = None
    recommends: list[str] | None = None
    requires: list[str] | None = None
    file: str | None = None
    binary: str | None = None
    provides: list[str] | None = None


SPECS = [
    RPM(name='dinginessentail', version='1.0'),
    RPM(name='dinginessentail', version='1.0', release='2', epoch=1),
    RPM(name='dinginessentail', version='1.1', epoch=1),
    RPM(name='dinginessentail-olive', version='1.0'),
    RPM(name='dinginessentail-olive', version='1.1'),
    RPM(name='multilib-dinginessentail', version='1.0', arch=['i686', 'x86_64']),
    RPM(name='multilib-dinginessentail', version='1.1', arch=['i686', 'x86_64']),
    RPM(name='landsidescalping', version='1.0',),
    RPM(name='landsidescalping', version='1.1',),
    RPM(name='dinginessentail-with-weak-dep', version='1.0', recommends=['dinginessentail-weak-dep']),
    RPM(name='dinginessentail-weak-dep', version='1.0',),
    RPM(name='noarchfake', version='1.0'),
    RPM(name='provides_foo_a', version='1.0', file='foo.gif'),
    RPM(name='provides_foo_b', version='1.0', file='foo.gif'),
    RPM(name='number-11-name', version='11.0',),
    RPM(name='number-11-name', version='11.1',),
    RPM(name='epochone', version='1.0', epoch=1),
    RPM(name='epochone', version='1.1', epoch=1),
    RPM(name='broken-a', version='1.2.3',),
    RPM(name='broken-a', version='1.2.3.4', requires=['dinginessentail-doesnotexist']),
    RPM(name='broken-a', version='1.2.4',),
    RPM(name='broken-a', version='2.0.0', requires=['dinginessentail-doesnotexist']),
    RPM(name='broken-b', version='1.0', requires=['broken-a = 1.2.3-1']),
    RPM(name='broken-c', version='1.0', requires=['broken-c = 1.2.4-1']),
    RPM(name='broken-d', version='1.0', requires=['broken-a']),
    RPM(name='provides-binary', version='1.0', arch=[expectedArch], binary='/usr/sbin/package-name'),
    RPM(name='package-name', version='1.0'),
    RPM(name='provides-package', version='1.0', provides=['provided-package']),
    RPM(name='provided-package', version='1.0'),
]


def create_repo():
    pkgs = []
    for spec in SPECS:
        pkg = SimpleRpmBuild(spec.name, spec.version, spec.release, spec.arch or ['noarch'])
        pkg.epoch = spec.epoch

        for requires in spec.requires or []:
            pkg.add_requires(requires)

        for recommend in spec.recommends or []:
            pkg.add_recommends(recommend)

        for provide in spec.provides or []:
            pkg.add_provides(provide)

        if spec.file:
            pkg.add_installed_file(
                "/" + spec.file,
                GeneratedSourceFile(
                    spec.file, make_gif()
                )
            )

        if spec.binary:
            pkg.add_simple_compilation(installPath=spec.binary)

        pkgs.append(pkg)

    repo = YumRepoBuild(pkgs)
    repo.make('noarch', 'i686', 'x86_64', expectedArch)

    for pkg in pkgs:
        pkg.clean()

    return repo.repoDir


def main():
    module = AnsibleModule(
        argument_spec={
            'tempdir': {'type': 'path'},
        }
    )

    if not HAS_RPMFLUFF:
        system_interpreters = ['/usr/libexec/platform-python', '/usr/bin/python3', '/usr/bin/python']

        interpreter = probe_interpreters_for_module(system_interpreters, 'rpmfluff')

        if not interpreter or has_respawned():
            module.fail_json('unable to find rpmfluff; tried {0}'.format(system_interpreters))

        respawn_module(interpreter)

    tempdir = module.params['tempdir']

    # Save current temp dir so we can set it back later
    original_tempdir = tempfile.tempdir
    tempfile.tempdir = tempdir

    try:
        repo_dir = create_repo()
    finally:
        tempfile.tempdir = original_tempdir

    module.exit_json(repo_dir=repo_dir, tmpfile=tempfile.gettempdir())


if __name__ == "__main__":
    main()



from __future__ import annotations as _annotations

import inspect as _inspect
import typing as _t


def caller_frame() -> _inspect.FrameInfo | None:
    """Return the caller stack frame, skipping any marked with the `_skip_stackwalk` local."""
    _skip_stackwalk = True

    return next(iter_stack(), None)


def iter_stack() -> _t.Generator[_inspect.FrameInfo]:
    """Iterate over stack frames, skipping any marked with the `_skip_stackwalk` local."""
    _skip_stackwalk = True

    for frame_info in _inspect.stack():
        if '_skip_stackwalk' in frame_info.frame.f_locals:
            continue

        yield frame_info



# Copyright: (c) 2017, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

from ansible.plugins.action import ActionBase


class ActionModule(ActionBase):

    def run(self, tmp=None, task_vars=None):
        del tmp  # tmp no longer has any effect

        try:
            self._task.args['_raw_params'] = self._task.args.pop('cmd')
        except KeyError:
            pass
        shell_action = self._shared_loader_obj.action_loader.get('ansible.legacy.shell',
                                                                 task=self._task,
                                                                 connection=self._connection,
                                                                 play_context=self._play_context,
                                                                 loader=self._loader,
                                                                 templar=self._templar,
                                                                 shared_loader_obj=self._shared_loader_obj)
        result = shell_action.run(task_vars=task_vars)
        result.pop('failed', None)
        return result



# -*- coding: utf-8 -*-
# Copyright: (c) 2018, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

DOCUMENTATION = """
    name: runas
    short_description: Run As user
    description:
        - This become plugin allows your remote/login user to execute commands as another user via the windows runas facility.
    author: ansible (@core)
    version_added: "2.8"
    options:
        become_user:
            description: User you 'become' to execute the task
            ini:
              - section: privilege_escalation
                key: become_user
              - section: runas_become_plugin
                key: user
            vars:
              - name: ansible_become_user
              - name: ansible_runas_user
            env:
              - name: ANSIBLE_BECOME_USER
              - name: ANSIBLE_RUNAS_USER
            keyword:
              - name: become_user
            required: True
        become_flags:
            description: Options to pass to runas, a space delimited list of k=v pairs
            default: ''
            ini:
              - section: privilege_escalation
                key: become_flags
              - section: runas_become_plugin
                key: flags
            vars:
              - name: ansible_become_flags
              - name: ansible_runas_flags
            env:
              - name: ANSIBLE_BECOME_FLAGS
              - name: ANSIBLE_RUNAS_FLAGS
            keyword:
              - name: become_flags
        become_pass:
            description: password
            ini:
              - section: runas_become_plugin
                key: password
            vars:
              - name: ansible_become_password
              - name: ansible_become_pass
              - name: ansible_runas_pass
            env:
              - name: ANSIBLE_BECOME_PASS
              - name: ANSIBLE_RUNAS_PASS
    notes:
        - runas is really implemented in the powershell module handler and as such can only be used with winrm connections.
        - This plugin ignores the 'become_exe' setting as it uses an API and not an executable.
        - The Secondary Logon service (seclogon) must be running to use runas
"""

from ansible.errors import AnsibleError
from ansible.parsing.splitter import split_args
from ansible.plugins.become import BecomeBase


class BecomeModule(BecomeBase):

    name = 'runas'

    def build_become_command(self, cmd, shell):
        # this is a noop, the 'real' runas is implemented
        # inside the windows powershell execution subsystem
        return cmd

    def _build_powershell_wrapper_action(self) -> tuple[str, dict[str, object], dict[str, object]]:
        # See ansible.executor.powershell.become_wrapper.ps1 for the
        # parameter names
        params = {
            'BecomeUser': self.get_option('become_user'),
        }
        secure_params = {}

        password = self.get_option('become_pass')
        if password:
            secure_params['BecomePassword'] = password

        flags = self.get_option('become_flags')
        if flags:
            split_flags = split_args(flags)
            for flag in split_flags:
                if '=' not in flag:
                    raise ValueError(f"become_flags entry '{flag}' is in an invalid format, must be a key=value pair")

                k, v = flag.split('=', 1)

                param_name, param_value = self._parse_flag(k, v)
                params[param_name] = param_value

        return 'become_wrapper.ps1', params, secure_params

    def _parse_flag(self, name: str, value: str) -> tuple[str, str]:
        logon_types = {
            'interactive': 'Interactive',
            'network': 'Network',
            'batch': 'Batch',
            'service': 'Service',
            'unlock': 'Unlock',
            'network_cleartext': 'NetworkCleartext',
            'new_credentials': 'NewCredentials',
        }
        logon_flags = {
            'none': 'None',
            'with_profile': 'WithProfile',
            'netcredentials_only': 'NetCredentialsOnly',
        }

        match name.lower():
            case 'logon_type':
                param_name = 'LogonType'
                if param_value := logon_types.get(value.lower(), None):
                    return param_name, param_value
                else:
                    raise AnsibleError(f"become_flags logon_type value '{value}' is not valid, valid values are: {', '.join(logon_types.keys())}")

            case 'logon_flags':
                param_name = 'LogonFlags'
                flags = value.split(',')

                param_values: list[str] = []
                for flag in flags:
                    if not flag:
                        continue

                    if flag_value := logon_flags.get(flag.lower(), None):
                        param_values.append(flag_value)
                    else:
                        raise AnsibleError(f"become_flags logon_flags value '{flag}' is not valid, valid values are: {', '.join(logon_flags.keys())}")

                return param_name, ", ".join(param_values)

            case _:
                raise AnsibleError(f"become_flags key '{name}' is not a valid runas flag, must be 'logon_type' or 'logon_flags'")



"""Provides compatibility with first-generation host delegation options in ansible-test."""

from __future__ import annotations

import argparse
import collections.abc as c
import dataclasses
import enum
import os
import types
import typing as t

from ..constants import (
    CONTROLLER_PYTHON_VERSIONS,
    SUPPORTED_PYTHON_VERSIONS,
)

from ..util import (
    ApplicationError,
    display,
    filter_args,
    sorted_versions,
)

from ..docker_util import (
    docker_available,
)

from ..completion import (
    docker_completion,
    remote_completion,
    filter_completion,
    windows_completion,
)

from ..host_configs import (
    ControllerConfig,
    ControllerHostConfig,
    DockerConfig,
    FallbackDetail,
    FallbackReason,
    HostConfig,
    HostContext,
    HostSettings,
    NativePythonConfig,
    NetworkInventoryConfig,
    NetworkRemoteConfig,
    OriginConfig,
    PosixRemoteConfig,
    VirtualPythonConfig,
    WindowsInventoryConfig,
    WindowsRemoteConfig,
)

from ..data import (
    data_context,
)


def filter_python(version: t.Optional[str], versions: t.Optional[c.Sequence[str]]) -> t.Optional[str]:
    """If a Python version is given and is in the given version list, return that Python version, otherwise return None."""
    return version if version in versions else None


def controller_python(version: t.Optional[str]) -> t.Optional[str]:
    """If a Python version is given and is supported by the controller, return that Python version, otherwise return None."""
    return filter_python(version, CONTROLLER_PYTHON_VERSIONS)


def get_fallback_remote_controller() -> str:
    """Return the remote fallback platform for the controller."""
    fallback = [item for name, item in filter_completion(remote_completion()).items() if item.controller_supported and name == "freebsd/latest"][0]
    return fallback.name


def get_option_name(name: str) -> str:
    """Return a command-line option name from the given option name."""
    if name == 'targets':
        name = 'target'

    return f'--{name.replace("_", "-")}'


class PythonVersionUnsupportedError(ApplicationError):
    """A Python version was requested for a context which does not support that version."""

    def __init__(self, context: str, version: str, versions: c.Iterable[str]) -> None:
        super().__init__(f'Python {version} is not supported by environment `{context}`. Supported Python version(s) are: {", ".join(versions)}')


class PythonVersionUnspecifiedError(ApplicationError):
    """A Python version was not specified for a context which is unknown, thus the Python version is unknown."""

    def __init__(self, context: str) -> None:
        super().__init__(
            f'Environment `{context}` is unknown. Use a predefined environment instead. '
            f'Alternatively, to use an unknown environment, use the `--python` option to specify a Python version.'
        )


class ControllerNotSupportedError(ApplicationError):
    """Option(s) were specified which do not provide support for the controller and would be ignored because they are irrelevant for the target."""

    def __init__(self, context: str) -> None:
        super().__init__(f'Environment `{context}` does not provide a Python version supported by the controller.')


class OptionsConflictError(ApplicationError):
    """Option(s) were specified which conflict with other options."""

    def __init__(self, first: c.Iterable[str], second: c.Iterable[str]) -> None:
        super().__init__(f'Options `{" ".join(first)}` cannot be combined with options `{" ".join(second)}`.')


@dataclasses.dataclass(frozen=True)
class LegacyHostOptions:
    """Legacy host options used prior to the availability of separate controller and target host configuration."""

    python: t.Optional[str] = None
    python_interpreter: t.Optional[str] = None
    local: t.Optional[bool] = None
    venv: t.Optional[bool] = None
    venv_system_site_packages: t.Optional[bool] = None
    remote: t.Optional[str] = None
    remote_provider: t.Optional[str] = None
    remote_arch: t.Optional[str] = None
    docker: t.Optional[str] = None
    docker_privileged: t.Optional[bool] = None
    docker_seccomp: t.Optional[str] = None
    docker_memory: t.Optional[int] = None
    windows: t.Optional[list[str]] = None
    platform: t.Optional[list[str]] = None
    platform_collection: t.Optional[list[tuple[str, str]]] = None
    platform_connection: t.Optional[list[tuple[str, str]]] = None
    inventory: t.Optional[str] = None

    @staticmethod
    def create(namespace: t.Union[argparse.Namespace, types.SimpleNamespace]) -> LegacyHostOptions:
        """Create legacy host options from the given namespace."""
        kwargs = {field.name: getattr(namespace, field.name, None) for field in dataclasses.fields(LegacyHostOptions)}

        if kwargs['python'] == 'default':
            kwargs['python'] = None

        return LegacyHostOptions(**kwargs)

    @staticmethod
    def purge_namespace(namespace: t.Union[argparse.Namespace, types.SimpleNamespace]) -> None:
        """Purge legacy host options fields from the given namespace."""
        for field in dataclasses.fields(LegacyHostOptions):
            if hasattr(namespace, field.name):
                delattr(namespace, field.name)

    @staticmethod
    def purge_args(args: list[str]) -> list[str]:
        """Purge legacy host options from the given command line arguments."""
        fields: tuple[dataclasses.Field, ...] = dataclasses.fields(LegacyHostOptions)
        filters: dict[str, int] = {get_option_name(field.name): 0 if field.type is t.Optional[bool] else 1 for field in fields}

        return filter_args(args, filters)

    def get_options_used(self) -> tuple[str, ...]:
        """Return a tuple of the command line options used."""
        fields: tuple[dataclasses.Field, ...] = dataclasses.fields(self)
        options = tuple(sorted(get_option_name(field.name) for field in fields if getattr(self, field.name)))
        return options


class TargetMode(enum.Enum):
    """Type of provisioning to use for the targets."""

    WINDOWS_INTEGRATION = enum.auto()  # windows-integration
    NETWORK_INTEGRATION = enum.auto()  # network-integration
    POSIX_INTEGRATION = enum.auto()  # integration
    SANITY = enum.auto()  # sanity
    UNITS = enum.auto()  # units
    SHELL = enum.auto()  # shell
    NO_TARGETS = enum.auto()  # coverage

    @property
    def one_host(self) -> bool:
        """Return True if only one host (the controller) should be used, otherwise return False."""
        return self in (TargetMode.SANITY, TargetMode.UNITS, TargetMode.NO_TARGETS)

    @property
    def no_fallback(self) -> bool:
        """Return True if no fallback is acceptable for the controller (due to options not applying to the target), otherwise return False."""
        return self in (TargetMode.WINDOWS_INTEGRATION, TargetMode.NETWORK_INTEGRATION, TargetMode.NO_TARGETS)

    @property
    def multiple_pythons(self) -> bool:
        """Return True if multiple Python versions are allowed, otherwise False."""
        return self in (TargetMode.SANITY, TargetMode.UNITS)

    @property
    def has_python(self) -> bool:
        """Return True if this mode uses Python, otherwise False."""
        return self in (TargetMode.POSIX_INTEGRATION, TargetMode.SANITY, TargetMode.UNITS, TargetMode.SHELL)


def convert_legacy_args(
    argv: list[str],
    args: t.Union[argparse.Namespace, types.SimpleNamespace],
    mode: TargetMode,
) -> HostSettings:
    """Convert pre-split host arguments in the given namespace to their split counterparts."""
    old_options = LegacyHostOptions.create(args)
    old_options.purge_namespace(args)

    new_options = [
        '--controller',
        '--target',
        '--target-python',
        '--target-posix',
        '--target-windows',
        '--target-network',
    ]

    used_old_options = old_options.get_options_used()
    used_new_options = [name for name in new_options if name in argv]

    if used_old_options:
        if used_new_options:
            raise OptionsConflictError(used_old_options, used_new_options)

        controller, targets, controller_fallback = get_legacy_host_config(mode, old_options)

        if controller_fallback:
            if mode.one_host:
                display.info(controller_fallback.message, verbosity=1)
            else:
                display.warning(controller_fallback.message)

        used_default_pythons = mode in (TargetMode.SANITY, TargetMode.UNITS) and not native_python(old_options)
    else:
        controller = args.controller or OriginConfig()
        controller_fallback = None

        if mode == TargetMode.NO_TARGETS:
            targets = []
            used_default_pythons = False
        elif args.targets:
            targets = args.targets
            used_default_pythons = False
        else:
            targets = default_targets(mode, controller)
            used_default_pythons = mode in (TargetMode.SANITY, TargetMode.UNITS)

    args.controller = controller
    args.targets = targets

    if used_default_pythons:
        control_targets = t.cast(list[ControllerConfig], targets)
        skipped_python_versions = sorted_versions(list(set(SUPPORTED_PYTHON_VERSIONS) - {target.python.version for target in control_targets}))
    else:
        skipped_python_versions = []

    filtered_args = old_options.purge_args(argv)
    filtered_args = filter_args(filtered_args, {name: 1 for name in new_options})

    host_settings = HostSettings(
        controller=controller,
        targets=targets,
        skipped_python_versions=skipped_python_versions,
        filtered_args=filtered_args,
        controller_fallback=controller_fallback,
    )

    return host_settings


def controller_targets(
    mode: TargetMode,
    options: LegacyHostOptions,
    controller: ControllerHostConfig,
) -> list[HostConfig]:
    """Return the configuration for controller targets."""
    python = native_python(options)

    targets: list[HostConfig]

    if python:
        targets = [ControllerConfig(python=python)]
    else:
        targets = default_targets(mode, controller)

    return targets


def native_python(options: LegacyHostOptions) -> t.Optional[NativePythonConfig]:
    """Return a NativePythonConfig for the given version if it is not None, otherwise return None."""
    if not options.python and not options.python_interpreter:
        return None

    return NativePythonConfig(version=options.python, path=options.python_interpreter)


def get_legacy_host_config(
    mode: TargetMode,
    options: LegacyHostOptions,
) -> tuple[ControllerHostConfig, list[HostConfig], t.Optional[FallbackDetail]]:
    """
    Returns controller and target host configs derived from the provided legacy host options.
    The goal is to match the original behavior, by using non-split testing whenever possible.
    When the options support the controller, use the options for the controller and use ControllerConfig for the targets.
    When the options do not support the controller, use the options for the targets and use a default controller config influenced by the options.
    """
    venv_fallback = 'venv/default'
    docker_fallback = 'default'
    remote_fallback = get_fallback_remote_controller()

    controller_fallback: t.Optional[tuple[str, str, FallbackReason]] = None

    controller: t.Optional[ControllerHostConfig]
    targets: list[HostConfig]

    if options.venv:
        if controller_python(options.python) or not options.python:
            controller = OriginConfig(python=VirtualPythonConfig(version=options.python or 'default', system_site_packages=options.venv_system_site_packages))
        else:
            controller_fallback = f'origin:python={venv_fallback}', f'--venv --python {options.python}', FallbackReason.PYTHON
            controller = OriginConfig(python=VirtualPythonConfig(version='default', system_site_packages=options.venv_system_site_packages))

        if mode in (TargetMode.SANITY, TargetMode.UNITS):
            python = native_python(options)

            if python:
                control_targets = [ControllerConfig(python=python)]
            else:
                control_targets = controller.get_default_targets(HostContext(controller_config=controller))

            # Target sanity tests either have no Python requirements or manage their own virtual environments.
            # Thus, there is no point in setting up virtual environments ahead of time for them.

            if mode == TargetMode.UNITS:
                targets = [ControllerConfig(python=VirtualPythonConfig(version=target.python.version, path=target.python.path,
                                                                       system_site_packages=options.venv_system_site_packages)) for target in control_targets]
            else:
                targets = t.cast(list[HostConfig], control_targets)
        else:
            targets = [ControllerConfig(python=VirtualPythonConfig(version=options.python or 'default',
                                                                   system_site_packages=options.venv_system_site_packages))]
    elif options.docker:
        docker_config = filter_completion(docker_completion()).get(options.docker)

        if docker_config:
            if options.python and options.python not in docker_config.supported_pythons:
                raise PythonVersionUnsupportedError(f'--docker {options.docker}', options.python, docker_config.supported_pythons)

            if docker_config.controller_supported:
                if controller_python(options.python) or not options.python:
                    controller = DockerConfig(name=docker_config.name, python=native_python(options),
                                              privileged=options.docker_privileged, seccomp=options.docker_seccomp, memory=options.docker_memory)
                    targets = controller_targets(mode, options, controller)
                else:
                    controller_fallback = f'docker:{options.docker}', f'--docker {options.docker} --python {options.python}', FallbackReason.PYTHON
                    controller = DockerConfig(name=docker_config.name)
                    targets = controller_targets(mode, options, controller)
            else:
                controller_fallback = f'docker:{docker_fallback}', f'--docker {options.docker}', FallbackReason.ENVIRONMENT
                controller = DockerConfig(name=docker_fallback)
                targets = [DockerConfig(name=docker_config.name, python=native_python(options),
                                        privileged=options.docker_privileged, seccomp=options.docker_seccomp, memory=options.docker_memory)]
        else:
            if not options.python:
                raise PythonVersionUnspecifiedError(f'--docker {options.docker}')

            if controller_python(options.python):
                controller = DockerConfig(name=options.docker, python=native_python(options),
                                          privileged=options.docker_privileged, seccomp=options.docker_seccomp, memory=options.docker_memory)
                targets = controller_targets(mode, options, controller)
            else:
                controller_fallback = f'docker:{docker_fallback}', f'--docker {options.docker} --python {options.python}', FallbackReason.PYTHON
                controller = DockerConfig(name=docker_fallback)
                targets = [DockerConfig(name=options.docker, python=native_python(options),
                                        privileged=options.docker_privileged, seccomp=options.docker_seccomp, memory=options.docker_memory)]
    elif options.remote:
        remote_config = filter_completion(remote_completion()).get(options.remote)
        context, reason = None, None

        if remote_config:
            if options.python and options.python not in remote_config.supported_pythons:
                raise PythonVersionUnsupportedError(f'--remote {options.remote}', options.python, remote_config.supported_pythons)

            if remote_config.controller_supported:
                if controller_python(options.python) or not options.python:
                    controller = PosixRemoteConfig(name=remote_config.name, python=native_python(options), provider=options.remote_provider,
                                                   arch=options.remote_arch)
                    targets = controller_targets(mode, options, controller)
                else:
                    controller_fallback = f'remote:{options.remote}', f'--remote {options.remote} --python {options.python}', FallbackReason.PYTHON
                    controller = PosixRemoteConfig(name=remote_config.name, provider=options.remote_provider, arch=options.remote_arch)
                    targets = controller_targets(mode, options, controller)
            else:
                context, reason = f'--remote {options.remote}', FallbackReason.ENVIRONMENT
                controller = None
                targets = [PosixRemoteConfig(name=remote_config.name, python=native_python(options), provider=options.remote_provider,
                                             arch=options.remote_arch)]
        elif mode == TargetMode.SHELL and options.remote.startswith('windows/'):
            if options.python and options.python not in CONTROLLER_PYTHON_VERSIONS:
                raise ControllerNotSupportedError(f'--python {options.python}')

            name = resolve_windows_names([options.remote.removeprefix("windows/")])[0]
            controller = OriginConfig(python=native_python(options))
            targets = [WindowsRemoteConfig(name=name, provider=options.remote_provider, arch=options.remote_arch)]
        else:
            if not options.python:
                raise PythonVersionUnspecifiedError(f'--remote {options.remote}')

            if controller_python(options.python):
                controller = PosixRemoteConfig(name=options.remote, python=native_python(options), provider=options.remote_provider, arch=options.remote_arch)
                targets = controller_targets(mode, options, controller)
            else:
                context, reason = f'--remote {options.remote} --python {options.python}', FallbackReason.PYTHON
                controller = None
                targets = [PosixRemoteConfig(name=options.remote, python=native_python(options), provider=options.remote_provider, arch=options.remote_arch)]

        if not controller:
            if docker_available():
                controller_fallback = f'docker:{docker_fallback}', context, reason
                controller = DockerConfig(name=docker_fallback)
            else:
                controller_fallback = f'remote:{remote_fallback}', context, reason
                controller = PosixRemoteConfig(name=remote_fallback)
    else:  # local/unspecified
        # There are several changes in behavior from the legacy implementation when using no delegation (or the `--local` option).
        # These changes are due to ansible-test now maintaining consistency between its own Python and that of controller Python subprocesses.
        #
        # 1) The `--python-interpreter` option (if different from sys.executable) now affects controller subprocesses and triggers re-execution of ansible-test.
        #    Previously this option was completely ignored except when used with the `--docker` or `--remote` options.
        # 2) The `--python` option now triggers re-execution of ansible-test if it differs from sys.version_info.
        #    Previously it affected Python subprocesses, but not ansible-test itself.

        if controller_python(options.python) or not options.python:
            controller = OriginConfig(python=native_python(options))
            targets = controller_targets(mode, options, controller)
        else:
            controller_fallback = 'origin:python=default', f'--python {options.python}', FallbackReason.PYTHON
            controller = OriginConfig()
            targets = controller_targets(mode, options, controller)

    if controller_fallback:
        controller_option, context, reason = controller_fallback

        if mode.no_fallback:
            raise ControllerNotSupportedError(context)

        fallback_detail = FallbackDetail(
            reason=reason,
            message=f'Using `--controller {controller_option}` since `{context}` does not support the controller.',
        )
    else:
        fallback_detail = None

    if mode.one_host and any(not isinstance(target, ControllerConfig) for target in targets):
        raise ControllerNotSupportedError(controller_fallback[1])

    if mode == TargetMode.NO_TARGETS:
        targets = []
    else:
        targets = handle_non_posix_targets(mode, options, targets)

    return controller, targets, fallback_detail


def handle_non_posix_targets(
    mode: TargetMode,
    options: LegacyHostOptions,
    targets: list[HostConfig],
) -> list[HostConfig]:
    """Return a list of non-POSIX targets if the target mode is non-POSIX."""
    if mode == TargetMode.WINDOWS_INTEGRATION:
        if options.windows:
            names = resolve_windows_names(options.windows)
            targets = [WindowsRemoteConfig(name=name, provider=options.remote_provider, arch=options.remote_arch) for name in names]
        else:
            targets = [WindowsInventoryConfig(path=options.inventory)]
    elif mode == TargetMode.NETWORK_INTEGRATION:
        if options.platform:
            network_targets = [NetworkRemoteConfig(name=platform, provider=options.remote_provider, arch=options.remote_arch) for platform in options.platform]

            for platform, collection in options.platform_collection or []:
                for entry in network_targets:
                    if entry.platform == platform:
                        entry.collection = collection

            for platform, connection in options.platform_connection or []:
                for entry in network_targets:
                    if entry.platform == platform:
                        entry.connection = connection

            targets = t.cast(list[HostConfig], network_targets)
        else:
            targets = [NetworkInventoryConfig(path=options.inventory)]

    return targets


def resolve_windows_names(versions: list[str]) -> list[str]:
    """Resolve a list of Windows versions into version names, resolving any aliases."""
    windows_completions = filter_completion(windows_completion())

    names = [f'windows/{version}' for version in versions]  # map versions to names
    names = [windows_completions[name].name if name in windows_completions else name for name in names]  # resolve aliases

    return names


def default_targets(
    mode: TargetMode,
    controller: ControllerHostConfig,
) -> list[HostConfig]:
    """Return a list of default targets for the given target mode."""
    targets: list[HostConfig]

    if mode == TargetMode.WINDOWS_INTEGRATION:
        targets = [WindowsInventoryConfig(path=os.path.abspath(os.path.join(data_context().content.integration_path, 'inventory.winrm')))]
    elif mode == TargetMode.NETWORK_INTEGRATION:
        targets = [NetworkInventoryConfig(path=os.path.abspath(os.path.join(data_context().content.integration_path, 'inventory.networking')))]
    elif mode.multiple_pythons:
        targets = t.cast(list[HostConfig], controller.get_default_targets(HostContext(controller_config=controller)))
    else:
        targets = [ControllerConfig()]

    return targets




from __future__ import annotations

import grp


def main():
    gids = [g.gr_gid for g in grp.getgrall()]

    # Start the gid numbering with 1
    # FreeBSD doesn't support the usage of gid 0, it doesn't fail (rc=0) but instead a number in the normal
    # range is picked.
    i = 1
    while True:
        if i not in gids:
            print(i)
            break
        i += 1


if __name__ == '__main__':
    main()



#!/usr/bin/python
from __future__ import annotations

import json
import sys

from ansible_collections.testns.testcoll.plugins.module_utils.moved_out_root import importme
from ..module_utils.formerly_testcoll_pkg import thing as movedthing  # pylint: disable=relative-beyond-top-level
from ..module_utils.formerly_testcoll_pkg.submod import thing as submodmovedthing  # pylint: disable=relative-beyond-top-level


def main():
    mu_result = importme()
    print(json.dumps(dict(changed=False, source='user', mu_result=mu_result, mu_result2=movedthing, mu_result3=submodmovedthing)))

    sys.exit()


if __name__ == '__main__':
    main()



from __future__ import annotations

import abc
import typing as t

from contextvars import ContextVar

from ansible.module_utils._internal._datatag import AnsibleTagHelper


class NotifiableAccessContextBase(metaclass=abc.ABCMeta):
    """Base class for a context manager that, when active, receives notification of managed access for types/tags in which it has registered an interest."""

    _type_interest: t.FrozenSet[type] = frozenset()
    """Set of types (including tag types) for which this context will be notified upon access."""

    _mask: t.ClassVar[bool] = False
    """When true, only the innermost (most recently created) context of this type will be notified."""

    def __enter__(self):
        # noinspection PyProtectedMember
        AnsibleAccessContext.current()._register_interest(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        # noinspection PyProtectedMember
        AnsibleAccessContext.current()._unregister_interest(self)
        return None

    @abc.abstractmethod
    def _notify(self, o: t.Any) -> t.Any:
        """Derived classes implement custom notification behavior when a registered type or tag is accessed."""


class AnsibleAccessContext:
    """
    Broker object for managed access registration and notification.
    Each thread or other logical callstack has a dedicated `AnsibleAccessContext` object with which `NotifiableAccessContext` objects can register interest.
    When a managed access occurs on an object, each active `NotifiableAccessContext` within the current callstack that has registered interest in that
    object's type or a tag present on it will be notified.
    """

    _contextvar: t.ClassVar[ContextVar[AnsibleAccessContext]] = ContextVar('AnsibleAccessContext')

    @staticmethod
    def current() -> AnsibleAccessContext:
        """Creates or retrieves an `AnsibleAccessContext` for the current logical callstack."""
        try:
            ctx: AnsibleAccessContext = AnsibleAccessContext._contextvar.get()
        except LookupError:
            # didn't exist; create it
            ctx = AnsibleAccessContext()
            AnsibleAccessContext._contextvar.set(ctx)  # we ignore the token, since this should live for the life of the thread/async ctx

        return ctx

    def __init__(self) -> None:
        self._notify_contexts: list[NotifiableAccessContextBase] = []

    def _register_interest(self, context: NotifiableAccessContextBase) -> None:
        self._notify_contexts.append(context)

    def _unregister_interest(self, context: NotifiableAccessContextBase) -> None:
        ctx = self._notify_contexts.pop()

        if ctx is not context:
            raise RuntimeError(f'Out-of-order context deactivation detected. Found {ctx} instead of {context}.')

    def access(self, value: t.Any) -> None:
        """Notify all contexts which have registered interest in the given value that it is being accessed."""
        if not self._notify_contexts:
            return

        value_types = AnsibleTagHelper.tag_types(value) | frozenset((type(value),))
        masked: set[type] = set()

        for ctx in reversed(self._notify_contexts):
            if ctx._mask:
                if (ctx_type := type(ctx)) in masked:
                    continue

                masked.add(ctx_type)

            # noinspection PyProtectedMember
            if ctx._type_interest.intersection(value_types):
                ctx._notify(value)



from __future__ import annotations

import keyword


def validate_collection_name(collection_name: object, name: str = 'collection_name') -> None:
    """Validate a collection name."""
    if not isinstance(collection_name, str):
        raise TypeError(f"{name} must be {str} instead of {type(collection_name)}")

    parts = collection_name.split('.')

    if len(parts) != 2 or not all(part.isidentifier() and not keyword.iskeyword(part) for part in parts):
        raise ValueError(f"{name} must consist of two non-keyword identifiers separated by '.'")



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os

from ansible.module_utils.common.text.converters import to_bytes
from ansible.playbook.attribute import NonInheritableFieldAttribute
from ansible.playbook.base import Base
from ansible.playbook.conditional import Conditional
from ansible.playbook.taggable import Taggable
from ansible.utils.collection_loader import AnsibleCollectionConfig
from ansible.utils.collection_loader._collection_finder import _get_collection_name_from_path, _get_collection_playbook_path
from ansible._internal._templating._engine import TemplateEngine
from ansible.errors import AnsibleError
from ansible import constants as C


class PlaybookInclude(Base, Conditional, Taggable):

    import_playbook = NonInheritableFieldAttribute(isa='string', required=True)

    _post_validate_object = True  # manually post_validate to get free arg validation/coercion

    def preprocess_data(self, ds):
        keys = {action for action in C._ACTION_IMPORT_PLAYBOOK if action in ds}

        if len(keys) != 1:
            raise AnsibleError(f'Found conflicting import_playbook actions: {", ".join(sorted(keys))}')

        key = next(iter(keys))

        ds['import_playbook'] = ds.pop(key)

        return ds

    @staticmethod
    def load(data, basedir, variable_manager=None, loader=None):
        return PlaybookInclude().load_data(ds=data, basedir=basedir, variable_manager=variable_manager, loader=loader)

    def load_data(self, ds, variable_manager=None, loader=None, basedir=None):
        """
        Overrides the base load_data(), as we're actually going to return a new
        Playbook() object rather than a PlaybookInclude object
        """

        # import here to avoid a dependency loop
        from ansible.playbook import Playbook
        from ansible.playbook.play import Play

        # first, we use the original parent method to correctly load the object
        # via the load_data/preprocess_data system we normally use for other
        # playbook objects
        new_obj = super(PlaybookInclude, self).load_data(ds, variable_manager, loader)

        all_vars = self.vars.copy()

        if variable_manager:
            all_vars |= variable_manager.get_vars()

        templar = TemplateEngine(loader=loader, variables=all_vars)

        new_obj.post_validate(templar)

        # then we use the object to load a Playbook
        pb = Playbook(loader=loader)

        file_name = new_obj.import_playbook

        # check for FQCN
        resource = _get_collection_playbook_path(file_name)

        if resource is not None:
            playbook = resource[1]
            playbook_collection = resource[2]
        else:
            # not FQCN try path
            playbook = file_name
            if not os.path.isabs(playbook):
                playbook = os.path.join(basedir, playbook)

            # might still be collection playbook
            playbook_collection = _get_collection_name_from_path(playbook)

        if playbook_collection:
            # it is a collection playbook, setup default collections
            AnsibleCollectionConfig.default_collection = playbook_collection
        else:
            # it is NOT a collection playbook, setup adjacent paths
            AnsibleCollectionConfig.playbook_paths.append(os.path.dirname(os.path.abspath(to_bytes(playbook, errors='surrogate_or_strict'))))
            # broken, see: https://github.com/ansible/ansible/issues/85357

        pb._load_playbook_data(file_name=playbook, variable_manager=variable_manager, vars=self.vars.copy())

        # finally, update each loaded playbook entry with any variables specified
        # on the included playbook and/or any tags which may have been set
        for entry in pb._entries:

            # conditional includes on a playbook need a marker to skip gathering
            if new_obj.when and isinstance(entry, Play):
                entry._included_conditional = new_obj.when[:]

            temp_vars = entry.vars | new_obj.vars
            param_tags = temp_vars.pop('tags', None)
            if param_tags is not None:
                entry.tags.extend(param_tags.split(','))
            entry.vars = temp_vars
            entry.tags = list(set(entry.tags).union(new_obj.tags))
            if entry._included_path is None:
                entry._included_path = os.path.dirname(playbook)

            # Check to see if we need to forward the conditionals on to the included
            # plays. If so, we can take a shortcut here and simply prepend them to
            # those attached to each block (if any)
            if new_obj.when:
                for task_block in (entry.pre_tasks + entry.roles + entry.tasks + entry.post_tasks):
                    task_block._when = new_obj.when[:] + task_block.when[:]

        return pb



# -*- coding: utf-8 -*-
# Copyright (c) 2018–2019, Sviatoslav Sydorenko <webknjaz@redhat.com>
# Simplified BSD License (see licenses/simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)
"""Test low-level utility functions from ``module_utils.common.collections``."""

from __future__ import annotations

import pytest

from collections.abc import Sequence
from ansible.module_utils.common.collections import ImmutableDict, is_iterable, is_sequence


class SeqStub:
    """Stub emulating a sequence type.

    >>> from collections.abc import Sequence
    >>> assert issubclass(SeqStub, Sequence)
    >>> assert isinstance(SeqStub(), Sequence)
    """


Sequence.register(SeqStub)


TEST_STRINGS = u'he', u'Україна', u'Česká republika'
TEST_STRINGS = TEST_STRINGS + tuple(s.encode('utf-8') for s in TEST_STRINGS)

TEST_ITEMS_NON_SEQUENCES: tuple = (
    {}, object(), frozenset(),
    4, 0.,
) + TEST_STRINGS

TEST_ITEMS_SEQUENCES: tuple = (
    [], (),
    SeqStub(),
    # Iterable effectively containing nested random data:
    TEST_ITEMS_NON_SEQUENCES,
)


@pytest.mark.parametrize('sequence_input', TEST_ITEMS_SEQUENCES)
def test_sequence_positive(sequence_input):
    """Test that non-string item sequences are identified correctly."""
    assert is_sequence(sequence_input)
    assert is_sequence(sequence_input, include_strings=False)


@pytest.mark.parametrize('non_sequence_input', TEST_ITEMS_NON_SEQUENCES)
def test_sequence_negative(non_sequence_input):
    """Test that non-sequences are identified correctly."""
    assert not is_sequence(non_sequence_input)


@pytest.mark.parametrize('string_input', TEST_STRINGS)
def test_sequence_string_types_with_strings(string_input):
    """Test that ``is_sequence`` can separate string and non-string."""
    assert is_sequence(string_input, include_strings=True)


@pytest.mark.parametrize('string_input', TEST_STRINGS)
def test_sequence_string_types_without_strings(string_input):
    """Test that ``is_sequence`` can separate string and non-string."""
    assert not is_sequence(string_input, include_strings=False)


@pytest.mark.parametrize(
    'seq',
    ([], (), {}, set(), frozenset()),
)
def test_iterable_positive(seq):
    assert is_iterable(seq)


@pytest.mark.parametrize(
    'seq', (object(), 5, 9.)
)
def test_iterable_negative(seq):
    assert not is_iterable(seq)


@pytest.mark.parametrize('string_input', TEST_STRINGS)
def test_iterable_including_strings(string_input):
    assert is_iterable(string_input, include_strings=True)


@pytest.mark.parametrize('string_input', TEST_STRINGS)
def test_iterable_excluding_strings(string_input):
    assert not is_iterable(string_input, include_strings=False)


class TestImmutableDict:
    def test_scalar(self):
        imdict = ImmutableDict({1: 2})
        assert imdict[1] == 2

    def test_string(self):
        imdict = ImmutableDict({u'café': u'くらとみ'})
        assert imdict[u'café'] == u'くらとみ'

    def test_container(self):
        imdict = ImmutableDict({(1, 2): ['1', '2']})
        assert imdict[(1, 2)] == ['1', '2']

    def test_from_tuples(self):
        imdict = ImmutableDict((('a', 1), ('b', 2)))
        assert frozenset(imdict.items()) == frozenset((('a', 1), ('b', 2)))

    def test_from_kwargs(self):
        imdict = ImmutableDict(a=1, b=2)
        assert frozenset(imdict.items()) == frozenset((('a', 1), ('b', 2)))

    def test_immutable(self):
        imdict = ImmutableDict({1: 2})

        expected_reason = r"^'ImmutableDict' object does not support item assignment$"

        with pytest.raises(TypeError, match=expected_reason):
            imdict[1] = 3

        with pytest.raises(TypeError, match=expected_reason):
            imdict[5] = 3

    def test_hashable(self):
        # ImmutableDict is hashable when all of its values are hashable
        imdict = ImmutableDict({u'café': u'くらとみ'})
        assert hash(imdict)

    def test_nonhashable(self):
        # ImmutableDict is unhashable when one of its values is unhashable
        imdict = ImmutableDict({u'café': u'くらとみ', 1: [1, 2]})

        expected_reason = r"^unhashable type: 'list'$"

        with pytest.raises(TypeError, match=expected_reason):
            hash(imdict)

    def test_len(self):
        imdict = ImmutableDict({1: 2, 'a': 'b'})
        assert len(imdict) == 2

    def test_repr(self):
        initial_data = {1: 2, 'a': 'b'}
        initial_data_repr = repr(initial_data)
        imdict = ImmutableDict(initial_data)
        actual_repr = repr(imdict)
        expected_repr = "ImmutableDict({0})".format(initial_data_repr)
        assert actual_repr == expected_repr



# -*- coding: utf-8 -*-
# Copyright: (c) 2022, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""Signature verification helpers."""
from __future__ import annotations

from ansible.errors import AnsibleError
from ansible.galaxy.user_agent import user_agent
from ansible.module_utils.urls import open_url

import contextlib
import inspect
import os
import subprocess
import typing as t

from dataclasses import dataclass, fields as dc_fields
from urllib.error import HTTPError, URLError

if t.TYPE_CHECKING:
    from ansible.utils.display import Display


def get_signature_from_source(source, display=None):  # type: (str, t.Optional[Display]) -> str
    if display is not None:
        display.vvvv(f"Using signature at {source}")
    try:
        with open_url(
            source,
            http_agent=user_agent(),
            validate_certs=True,
            follow_redirects='safe'
        ) as resp:
            signature = resp.read()
    except (HTTPError, URLError) as e:
        raise AnsibleError(
            f"Failed to get signature for collection verification from '{source}': {e}"
        ) from e

    return signature


def run_gpg_verify(
    manifest_file,  # type: str
    signature,  # type: str
    keyring,  # type: str
    display,  # type: Display
):  # type: (...) -> tuple[str, int]
    status_fd_read, status_fd_write = os.pipe()

    # running the gpg command will create the keyring if it does not exist
    remove_keybox = not os.path.exists(keyring)

    cmd = [
        'gpg',
        f'--status-fd={status_fd_write}',
        '--verify',
        '--batch',
        '--no-tty',
        '--no-default-keyring',
        f'--keyring={keyring}',
        '-',
        manifest_file,
    ]
    cmd_str = ' '.join(cmd)
    display.vvvv(f"Running command '{cmd}'")

    try:
        p = subprocess.Popen(
            cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            pass_fds=(status_fd_write,),
            encoding='utf8',
        )
    except (FileNotFoundError, subprocess.SubprocessError) as err:
        raise AnsibleError(
            f"Failed during GnuPG verification with command '{cmd_str}': {err}"
        ) from err
    else:
        stdout, stderr = p.communicate(input=signature)
    finally:
        os.close(status_fd_write)

    if remove_keybox:
        with contextlib.suppress(OSError):
            os.remove(keyring)

    with os.fdopen(status_fd_read) as f:
        stdout = f.read()
        display.vvvv(
            f"stdout: \n{stdout}\nstderr: \n{stderr}\n(exit code {p.returncode})"
        )
        return stdout, p.returncode


def parse_gpg_errors(status_out):  # type: (str) -> t.Iterator[GpgBaseError]
    for line in status_out.splitlines():
        if not line:
            continue
        try:
            _dummy, status, remainder = line.split(maxsplit=2)
        except ValueError:
            _dummy, status = line.split(maxsplit=1)
            remainder = None

        try:
            cls = GPG_ERROR_MAP[status]
        except KeyError:
            continue

        fields = [status]
        if remainder:
            fields.extend(
                remainder.split(
                    None,
                    len(dc_fields(cls)) - 2
                )
            )

        yield cls(*fields)


@dataclass(frozen=True, slots=True)
class GpgBaseError(Exception):
    status: str

    @classmethod
    def get_gpg_error_description(cls) -> str:
        """Return the current class description."""
        return ' '.join(cls.__doc__.split())

    def __post_init__(self):
        for field_name, field_type in inspect.get_annotations(type(self), eval_str=True).items():
            super(GpgBaseError, self).__setattr__(field_name, field_type(getattr(self, field_name)))


@dataclass(frozen=True, slots=True)
class GpgExpSig(GpgBaseError):
    """The signature with the keyid is good, but the signature is expired."""
    keyid: str
    username: str


@dataclass(frozen=True, slots=True)
class GpgExpKeySig(GpgBaseError):
    """The signature with the keyid is good, but the signature was made by an expired key."""
    keyid: str
    username: str


@dataclass(frozen=True, slots=True)
class GpgRevKeySig(GpgBaseError):
    """The signature with the keyid is good, but the signature was made by a revoked key."""
    keyid: str
    username: str


@dataclass(frozen=True, slots=True)
class GpgBadSig(GpgBaseError):
    """The signature with the keyid has not been verified okay."""
    keyid: str
    username: str


@dataclass(frozen=True, slots=True)
class GpgErrSig(GpgBaseError):
    """"It was not possible to check the signature.  This may be caused by
    a missing public key or an unsupported algorithm.  A RC of 4
    indicates unknown algorithm, a 9 indicates a missing public
    key.
    """
    keyid: str
    pkalgo: int
    hashalgo: int
    sig_class: str
    time: int
    rc: int
    fpr: str


@dataclass(frozen=True, slots=True)
class GpgNoPubkey(GpgBaseError):
    """The public key is not available."""
    keyid: str


@dataclass(frozen=True, slots=True)
class GpgMissingPassPhrase(GpgBaseError):
    """No passphrase was supplied."""


@dataclass(frozen=True, slots=True)
class GpgBadPassphrase(GpgBaseError):
    """The supplied passphrase was wrong or not given."""
    keyid: str


@dataclass(frozen=True, slots=True)
class GpgNoData(GpgBaseError):
    """No data has been found.  Codes for WHAT are:
    - 1 :: No armored data.
    - 2 :: Expected a packet but did not find one.
    - 3 :: Invalid packet found, this may indicate a non OpenPGP
           message.
    - 4 :: Signature expected but not found.
    """
    what: str


@dataclass(frozen=True, slots=True)
class GpgUnexpected(GpgBaseError):
    """No data has been found.  Codes for WHAT are:
    - 1 :: No armored data.
    - 2 :: Expected a packet but did not find one.
    - 3 :: Invalid packet found, this may indicate a non OpenPGP
           message.
    - 4 :: Signature expected but not found.
    """
    what: str


@dataclass(frozen=True, slots=True)
class GpgError(GpgBaseError):
    """This is a generic error status message, it might be followed by error location specific data."""
    location: str
    code: int
    more: str = ""


@dataclass(frozen=True, slots=True)
class GpgFailure(GpgBaseError):
    """This is the counterpart to SUCCESS and used to indicate a program failure."""
    location: str
    code: int


@dataclass(frozen=True, slots=True)
class GpgBadArmor(GpgBaseError):
    """The ASCII armor is corrupted."""


@dataclass(frozen=True, slots=True)
class GpgKeyExpired(GpgBaseError):
    """The key has expired."""
    timestamp: int


@dataclass(frozen=True, slots=True)
class GpgKeyRevoked(GpgBaseError):
    """The used key has been revoked by its owner."""


@dataclass(frozen=True, slots=True)
class GpgNoSecKey(GpgBaseError):
    """The secret key is not available."""
    keyid: str


GPG_ERROR_MAP = {
    'EXPSIG': GpgExpSig,
    'EXPKEYSIG': GpgExpKeySig,
    'REVKEYSIG': GpgRevKeySig,
    'BADSIG': GpgBadSig,
    'ERRSIG': GpgErrSig,
    'NO_PUBKEY': GpgNoPubkey,
    'MISSING_PASSPHRASE': GpgMissingPassPhrase,
    'BAD_PASSPHRASE': GpgBadPassphrase,
    'NODATA': GpgNoData,
    'UNEXPECTED': GpgUnexpected,
    'ERROR': GpgError,
    'FAILURE': GpgFailure,
    'BADARMOR': GpgBadArmor,
    'KEYEXPIRED': GpgKeyExpired,
    'KEYREVOKED': GpgKeyRevoked,
    'NO_SECKEY': GpgNoSecKey,
}



#!/usr/bin/python
from __future__ import annotations

import json
import sys

from ansible.module_utils.formerly_core import thingtocall


def main():
    mu_result = thingtocall()
    print(json.dumps(dict(changed=False, source='user', mu_result=mu_result)))

    sys.exit()


if __name__ == '__main__':
    main()
